"""A collection of basic utility functions for control applications. In particular, it
contains functions for solving the LQR problems in continuous- and discrete-time,
discretization methods such as Runge-Kutta 4, and functions to build Control Barrier
Functions.

Some inspiration was drawn from `MPCtools <https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py>`_.
"""

from collections.abc import Iterable as _Iterable
from typing import Callable, Optional

import casadi as cs
import numpy as np
import numpy.typing as npt
from scipy.linalg import solve_continuous_are as _solve_continuous_are
from scipy.linalg import solve_discrete_are as _solve_discrete_are

from .math import SymOrNumType, SymType
from .math import dual_norm as _dual_norm
from .math import lie_derivative as _lie_derivative


def lqr(
    A: npt.NDArray[np.floating],
    B: npt.NDArray[np.floating],
    Q: npt.NDArray[np.floating],
    R: npt.NDArray[np.floating],
    M: Optional[npt.NDArray[np.floating]] = None,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    r"""Computes the solution to the continuous-time LQR problem.

    The LQR problem is to solve the following optimization problem

    .. math::
        \min_{u} \int_{0}^{\infty} x^\top Q x + u^\top R u + 2 x^\top M u

    for the linear time-invariant continuous-time system

    .. math:: \dot{x} = A x + B u.

    The (famous) solution takes the form of a state feedback law

    .. math:: u = -K x

    with a quadratic cost-to-go function

    .. math:: V(x) = x^\top P x.

    The function returns the optimal state feedback matrix :math:`K` and the quadratic
    terminal cost-to-go matrix :math:`P`. If not provided, ``M`` is assumed to be zero.

    Parameters
    ----------
    A : array
        State matrix.
    B : array
        Control input matrix.
    Q : array
        State weighting matrix.
    R : array
        Control input weighting matrix.
    M : array, optional
        Mixed state-input weighting matrix, by default ``None``.

    Returns
    -------
    tuple of two arrays
        Returns the optimal state feedback matrix :math:`K` and the quadratic terminal
        cost-to-go matrix :math:`P`.
    """
    P = _solve_continuous_are(A, B, Q, R, s=M)
    rhs = B.T.dot(P) if M is None else B.T.dot(P) + M.T
    K = np.linalg.solve(R, rhs)
    return K, P


def dlqr(
    A: npt.NDArray[np.floating],
    B: npt.NDArray[np.floating],
    Q: npt.NDArray[np.floating],
    R: npt.NDArray[np.floating],
    M: Optional[npt.NDArray[np.floating]] = None,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    r"""Computes the solution to the discrete-time LQR problem.

    The LQR problem is to solve the following optimization problem

    .. math::
        \min_{u} \sum_{t=0}^{\infty} x_t^\top Q x_t + u_t^\top R u_t + 2 x_t^\top M u_t

    for the linear time-invariant discrete-time system

    .. math:: x_{t+1} = A x_t + B u_t.

    The (famous) solution takes the form of a state feedback law

    .. math:: u_t = -K x_t

    with a quadratic cost-to-go function

    .. math:: V(x_t) = x_t^\top P x_t.

    The function returns the optimal state feedback matrix :math:`K` and the quadratic
    terminal cost-to-go matrix :math:`P`. If not provided, ``M`` is assumed to be zero.

    Parameters
    ----------
    A : array
        State matrix.
    B : array
        Control input matrix.
    Q : array
        State weighting matrix.
    R : array
        Control input weighting matrix.
    M : array, optional
        Mixed state-input weighting matrix, by default ``None``.

    Returns
    -------
    tuple of two arrays
        Returns the optimal state feedback matrix :math:`K` and the quadratic terminal
        cost-to-go matrix :math:`P`.
    """
    P = _solve_discrete_are(A, B, Q, R, s=M)
    rhs = B.T.dot(P).dot(A) if M is None else B.T.dot(P).dot(A) + M.T
    K = np.linalg.solve(B.T.dot(P).dot(B) + R, rhs)
    return K, P


def rk4(
    f: Callable[[SymOrNumType], SymOrNumType],
    x0: SymOrNumType,
    dt: float = 1,
    M: int = 1,
) -> SymOrNumType:
    r"""Computes the Runge-Kutta 4 integration of the given function ``f`` with initial
    state ``x0``.

    In mathematical terms, given a continuous-time dynamics function
    :math:`\dot{x} = f(x)`, this method returns a discretized version of the dynamics
    :math:`x_{k+1} = f_d(x_k)` using the Runge-Kutta 4 method.

    Parameters
    ----------
    f : Callable[[casadi or array], casadi or array]
        A function that takes a state as input and returns the derivative of the state,
        i.e., continuous-time dynamics.
    x0 : casadi or array
        The initial state. Must be compatible as an argument to ``f``.
    dt : float, optional
        The discretization timestep, by default ``1``.
    M : int, optional
        How many RK4 steps to take in one ``dt`` interval, by default ``1``.

    Returns
    -------
    new state : casadi or array
        The new state after ``dt`` time, according to the discretization.
    """
    dt /= M
    x = x0
    for _ in range(M):
        k1 = f(x)
        k2 = f(x + k1 * dt / 2)
        k3 = f(x + k2 * dt / 2)
        k4 = f(x + k3 * dt)
        x = x + (k1 + 2 * k2 + 2 * k3 + k4) * dt / 6
    return x


def cbf(
    h: Callable[[SymType], SymType],
    x: SymType,
    u: SymType,
    dynamics: Callable[[SymType, SymType], SymType],
    alphas: _Iterable[Callable[[SymType], SymType]],
) -> SymType:
    r"""Continuous-time Control Barrier Function (CBF) for the given constraint ``h``
    and system with dynamics ``dynamics``. This method constructs a CBF for the
    constraint :math:`h(x) \geq 0` using the given system's dynamics
    :math:`\dot{x} = f(x, u)`. Here, :math:`\dot{x}` is the time derivative of the state
    after applying control input :math:`u`, and :math:`f` is the dynamics function
    ``dynamics``.

    The method can also compute a High-Order (HO) CBF by passing more than one class
    :math:`\mathcal{K}` functions ``alphas``.

    As per [1]_, the HO-CBF :math:`\phi_m` of degree :math:`m` is recursively found as

    .. math::
        \phi_m(x) = \dot{\phi}_{m-1}(x) + \alpha_m(\phi_{m-1}(x))

    and should be imposed as the constraint :math:`\phi_m(x) \geq 0`.

    Parameters
    ----------
    h : callable
        The constraint function for which to build the CBF. It must be of the signature
        :math:`x \rightarrow h(x)`.
    x : casadi SX or MX
        The state vector variable :math:`x`.
    u : casadi SX or MX
        The control input vector variable :math:`u`.
    dynamics : callable
        The dynamics function :math:`f` with signature :math:`x,u \rightarrow f(x, u)`.
    alphas : iterable of callables
        An iterable of class :math:`\mathcal{K}` functions :math:`\alpha_m` for
        the HO-CBF. The length of the iterable determines the degree of the HO-CBF.

    Returns
    -------
    casadi SX or MX
        Returns the HO-CBF function :math:`\phi_m` as a symbolic variable that is
        function of the provided ``x`` and ``u``.

    References
    ----------
    .. [1] Wei Xiao and Calin Belta. High-order control barrier functions. *IEEE
           Transactions on Automatic Control*, 67(7), 3655-3662, 2021.

    Examples
    --------
    >>> import casadi as cs
    >>> A = cs.SX.sym("A", 2, 2)
    >>> B = cs.SX.sym("B", 2, 1)
    >>> x = cs.SX.sym("x", A.shape[0], 1)
    >>> u = cs.SX.sym("u", B.shape[1], 1)
    >>> dynamics = lambda x, u: A @ x + B @ u
    >>> M = cs.SX.sym("M")
    >>> c = cs.SX.sym("c")
    >>> gamma = cs.SX.sym("gamma")
    >>> alphas = [lambda z: gamma * z]
    >>> h = lambda x: M - c * x[0]  # >= 0
    >>> cbf = cbf(h, x, u, dynamics, alphas)
    >>> print(cbf)
    """
    x_dot = dynamics(x, u)
    phi = h(x)
    for alpha in alphas:
        phi = _lie_derivative(phi, x, x_dot) + alpha(phi)
    # cs.depends_on(phi, u)
    # phi.which_depends("u", [name], 2, True)[0]
    return phi


def dcbf(
    h: Callable[[SymType], SymType],
    x: SymType,
    u: SymType,
    dynamics: Callable[[SymType, SymType], SymType],
    alphas: _Iterable[Callable[[SymType], SymType]],
) -> cs.Function:
    r"""Discrete-time Control Barrier Function (DCBF) for the given constraint ``h`` and
    system with dynamics ``dynamics``. This method constructs a DCBF for the constraint
    :math:`h(x) \geq 0` using the given system's dynamics :math:`x_{+} = f(x, u)`. Here,
    :math:`x_{+}` is the next state after applying control input :math:`u`, and
    :math:`f` is the dynamics function ``dynamics``.

    The method can also compute a High-Order (HO) DCBF by passing more than one class
    :math:`\mathcal{K}` functions ``alphas``.

    As per [1]_, the HO-DCBF :math:`\phi_m` of degree :math:`m` is recursively found as

    .. math::
        \phi_m(x_k) = \phi_{m-1}(x_{k+1}) - \phi_{m-1}(x_k) + \alpha_m(\phi_{m-1}(x_k))

    and should be imposed as the constraint :math:`\phi_m(x_k) \geq 0`.

    Parameters
    ----------
    h : callable
        The constraint function for which to build the DCBF. It must be of the signature
        :math:`x \rightarrow h(x)`.
    x : casadi SX or MX
        The state vector variable :math:`x`.
    u : casadi SX or MX
        The control input vector variable :math:`u`.
    dynamics : callable
        The dynamics function :math:`f` with signature :math:`x,u \rightarrow f(x, u)`.
    alphas : iterable of callables
        An iterable of class :math:`\mathcal{K}` functions :math:`\alpha_m` for
        the HO-DCBF. The length of the iterable determines the degree of the HO-DCBF.

    Returns
    -------
    casadi SX or MX
        Returns the HO-DCBF function :math:`\phi_m` as a symbolic variable that is
        function of the provided ``x`` and ``u``.

    References
    ----------
    .. [1] Yuhang Xiong, Di-Hua Zhai, Mahdi Tavakoli, Yuanqing Xia. Discrete-time
       control barrier function: High-order case and adaptive case. *IEEE Transactions
       on Cybernetics*, 53(5), 3231-3239, 2022.

    Examples
    --------
    >>> import casadi as cs
    >>> A = cs.SX.sym("A", 2, 2)
    >>> B = cs.SX.sym("B", 2, 1)
    >>> x = cs.SX.sym("x", A.shape[0], 1)
    >>> u = cs.SX.sym("u", B.shape[1], 1)
    >>> dynamics = lambda x, u: A @ x + B @ u
    >>> M = cs.SX.sym("M")
    >>> c = cs.SX.sym("c")
    >>> gamma = cs.SX.sym("gamma")
    >>> alphas = [lambda z: gamma * z]
    >>> h = lambda x: M - c * x[0]  # >= 0
    >>> cbf = dcbf(h, x, u, dynamics, alphas)
    >>> print(cbf)
    """
    x_next = dynamics(x, u)
    phi = h(x)
    for alpha in alphas:
        phi_next = cs.substitute(phi, x, x_next)
        phi = phi_next - phi + alpha(phi)
    return phi


def iccbf(
    h: Callable[[SymType], SymType],
    x: SymType,
    u: SymType,
    dynamics_f_and_g: Callable[[SymType], tuple[SymType, SymType]],
    alphas: _Iterable[Callable[[SymType], SymType]],
    norm: float,
    bound: float = 1.0,
) -> cs.Function:
    r"""Continuous-time Input Constrained Control Barrier Function (ICCBF) for the given
    constraint ``h`` and system with dynamics ``dynamics_f_and_g`` subject to
    norm-bounded control action. This method constructs an ICCBF for the constraint
    :math:`h(x) \geq 0` using the given system's control-affine dynamics
    :math:`\dot{x} = f(x) + g(x) u`. Here, :math:`\dot{x}` is the time derivative of the
    state after applying control input :math:`u`, and :math:`f` and :math:`g` are
    returned by the dynamics function ``dynamics_f_and_g``. The ICCBF takes into account
    the norm-bounded control input :math:`|| u ||_p \leq b`, where the norm power
    :math:`p` and bound :math:`b` are specified as ``norm`` and ``bound``.

    The method can also compute a High-Order (HO) ICCBF by passing more than one class
    :math:`\mathcal{K}` functions ``alphas``.

    As per [1]_, the HO-ICCBF :math:`\phi_m` of degree :math:`m` is recursively found as

    .. math::
        \phi_m(x) = L_f \phi_{m-1}(x)
                    + \inf_{u \in \mathcal{U}} \left\{ L_g \phi_{m-1}(x) u \right\}
                    + \alpha_m(\phi_{m-1}(x))

    and should be imposed as the constraint :math:`\phi_m(x) \geq 0`.

    Parameters
    ----------
    h : callable
        The constraint function for which to build the ICCBF. It must be of the
        signature :math:`x \rightarrow h(x)`.
    x : casadi SX or MX
        The state vector variable :math:`x`.
    u : casadi SX or MX
        The control input vector variable :math:`u`.
    dynamics_f_and_g : callable
        A callable computing, for the given state, the dynamics components :math:`f(x)`
        and :math:`g(x)`, i.e., the signature is :math:`x \rightarrow f(x), g(x)`.
    alphas : iterable of callables
        An iterable of class :math:`\mathcal{K}` functions :math:`\alpha_m` for
        the HO-ICCBF. The length of the iterable determines the degree of the HO-ICCBF.
    norm : float
        The norm power :math:`p` for the control input norm constraint
        :math:`||u||_p \le b`, with :math:`1 \le p \le \infty`.
    bound : float, optional
        The bound :math:`b` for the control input norm constraint :math:`||u||_p \le b`,
        by default ``1.0``. Must be larger than zero.

    Returns
    -------
    casadi SX or MX
        Returns the HO-ICCBF function :math:`\phi_m` as a symbolic variable that is
        function of the provided ``x`` and ``u``.

    References
    ----------
    .. [1] Devansh R. Agrawal and Dimitra Panagou. Safe control synthesis via input
           constrained control barrier functions. In *60th IEEE Conference on Decision
           and Control (CDC)*, 6113-6118, 2021.

    Examples
    --------
    >>> import casadi as cs
    >>> A = cs.SX.sym("A", 2, 2)
    >>> B = cs.SX.sym("B", 2, 1)
    >>> x = cs.SX.sym("x", A.shape[0], 1)
    >>> u = cs.SX.sym("u", B.shape[1], 1)
    >>> dynamics_f_and_g = lambda x: (A @ x, B)
    >>> M = cs.SX.sym("M")
    >>> c = cs.SX.sym("c")
    >>> gamma = cs.SX.sym("gamma")
    >>> alphas = [lambda z: gamma * z]
    >>> h = lambda x: M - c * x[0]  # >= 0
    >>> cbf = iccbf(h, x, u, dynamics_f_and_g, alphas, norm=2, bound=0.5)
    >>> print(cbf)
    """
    # find the dual norm of the given norm
    y = cs.SX.sym("y", u.shape[0], 1)
    dual_norm_func = cs.Function("dual_norm", (y,), (_dual_norm(y, norm),))

    # continue from here as usual
    f, g = dynamics_f_and_g(x)
    phi = h(x)
    for alpha in alphas:
        Lf_phi = _lie_derivative(phi, x, f)
        Lg_phi = _lie_derivative(phi, x, g)
        Lg_phi_u_sup = bound * dual_norm_func(Lg_phi)
        alpha_phi = alpha(phi)
        phi = Lf_phi - Lg_phi_u_sup + alpha_phi
    return Lf_phi + Lg_phi * u + alpha_phi  # skip infimum in last iteration



"""A module with wrappers for agents and environments.

In :mod:`gymnasium`, wrappers are a common way to modify the behavior of environments to
specific needs; see :class:`gymnasium.Wrapper`. In this module, we provide new wrappers
for environments, as well as define our own :class:`mpcrl.wrappers_agents.Wrapper` class
to wrap also (learning and non-) agents.

Submodules
==========

.. autosummary::
   :toctree: generated
   :template: module.rst
   :caption: Submodules

   agents
   envs
"""

__all__ = ["wrappers_agents", "wrappers_envs"]

from . import agents as wrappers_agents
from . import envs as wrappers_envs



"""The update strategy is likely to be one of the most important aspects that a designer
has to consider when training a Reinforcement Learning agent. When instantiating an
agent, via :class:`UpdateStrategy`, the user can specify when and with which frequency
to update the agent's MPC parametrization (e.g., at the end of every training episode,
or every ``N`` time steps), as well as the number of updates to skip at the beginning
(in case we need to wait for experience buffers to properly fill first). See
:ref:`user_guide_updating` for a more thorough explanation."""

from collections.abc import Iterator
from itertools import chain, repeat
from typing import Literal

from ..util.iters import bool_cycle


class UpdateStrategy:
    """A class holding information on the update strategy to be used by the learning
    algorithm.

    Parameters
    ----------
    frequency : int
        Frequency at which, each time the hook is called, an update should be
        carried out.
    skip_first : int, optional
        Skips the first ``skip_first`` updates. By default ``0``, so no update is
        skipped. This is useful when, e.g., the agent has to wait for the experience
        buffer to be filled before starting to update.
    hook : {"on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook, i.e., when to check if an update is due
        according to the given frequency. The options are:

        - ``"on_episode_end"`` checks if an update is due  after each episode ends
        - ``"on_timestep_end"`` checks for an update after each simulation's time step.

        By default, ``"on_timestep_end"`` is selected.
    """

    def __init__(
        self,
        frequency: int,
        hook: Literal["on_episode_end", "on_timestep_end"] = "on_timestep_end",
        skip_first: int = 0,
    ) -> None:
        self.frequency = frequency
        self.hook = hook
        self._update_cycle = chain(
            repeat(False, skip_first * frequency), bool_cycle(frequency)
        )

    def can_update(self) -> bool:
        """Returns whether an update must be carried out now, at the current instant,
        according to the specified strategy.

        Notes
        -----
        This methods steps the internal iterators to check whether an update is due with
        :func:`next`. This means that calling this method has a side effect on the state
        of these iterators, and calling immediately again can result in a different
        outcome.

        Returns
        -------
        bool
            ``True`` if the agent should update according to this strategy; otherwise,
            ``False``.
        """
        return next(self._update_cycle)

    def __iter__(self) -> Iterator[bool]:
        return self._update_cycle

    def __next__(self) -> bool:
        return next(self._update_cycle)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(frequency={self.frequency},hook={self.hook})"

    def __str__(self) -> str:
        return self.__repr__()



import sys
from collections.abc import Collection, Iterator
from typing import Callable, Generic, Literal, Optional, SupportsFloat, Union

import casadi as cs
import numba as nb
import numpy as np
import numpy.typing as npt
from csnlp.wrappers import Mpc, NlpSensitivity
from gymnasium import Env

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from ..core.experience import ExperienceReplay
from ..core.exploration import ExplorationStrategy, NoExploration
from ..core.parameters import LearnableParametersDict
from ..core.update import UpdateStrategy
from ..core.warmstart import WarmStartStrategy
from ..optim.gradient_based_optimizer import GradientBasedOptimizer
from ..util.math import monomials_basis_function
from .common.agent import ActType, ObsType, SymType
from .common.rl_learning_agent import LrType, RlLearningAgent

ExpType: TypeAlias = tuple[
    npt.NDArray[np.floating],  # rollout's costs
    npt.NDArray[np.floating],  # rollout's state feature vectors Phi(s)
    npt.NDArray[np.floating],  # rollout's Psi(s,a)
    npt.NDArray[np.floating],  # rollout's gradient of policy w.r.t. theta
    npt.NDArray[np.floating],  # rollout's CAFA weight v
]


class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, LrType]):
    r"""Least-Squares Temporal Difference (LSTD) Deterministic Policy Gradient (DPG)
    agent, as introduced in :cite:`gros_reinforcement_2021` as its stochastic
    counterpart, and refined in :cite:`gros_towards_2019`. An application can be found
    in :cite:`cai_mpcbased_2021`.

    The DPG agent uses an MPC controller as policy provider and function approximation,
    and adjusts its parametrization according to the temporal-difference error, with the
    goal of improving the policy, in a direct fashion by estimating the gradient of the
    policy and descending in its direction.

    Parameters
    ----------
    mpc : :class:`csnlp.wrappers.Mpc`
        The MPC controller used as policy provider by this agent. The instance is
        modified in place to create the approximations of the state function
        :math:`V_\theta(s)` and action value function :math:`Q_\theta(s,a)`, so it is
        recommended not to modify it further after initialization of the agent.
        Moreover, some parameter and constraint names will need to be created, so an
        error is thrown if these names are already in use in the mpc.
    update_strategy : UpdateStrategy or int
        The strategy used to decide which frequency to update the mpc parameters with.
        If an ``int`` is passed, then the default strategy that updates every ``n``
        episodes is used (where ``n`` is the argument passed); otherwise, an instance
        of :class:`core.update.UpdateStrategy` can be passed to specify the desired
        strategy in more details.
    discount_factor : float
        In RL, the factor that discounts future rewards in favor of immediate rewards.
        Usually denoted as :math:`\gamma`. It should satisfy :math:`\gamma \in (0, 1]`.
    optimizer : GradientBasedOptimizer
        A gradient-based optimizer (e.g., :class:`optim.GradientDescent`) to
        compute the updates of the learnable parameters, based on the current
        gradient-based RL algorithm.
    learnable_parameters : :class:`core.parameters.LearnableParametersDict`
        A special dict containing the learnable parameters of the MPC (usually referred
        to as :math:`\theta`), together with their bounds and values. This dict is
        complementary to :attr:`fixed_parameters`, which contains the MPC parameters
        that are not learnt by the agent.
    exploration : :class:`core.exploration.ExplorationStrategy`
        Exploration strategy for inducing exploration in the online MPC policy. It is
        mandatory for DPG agents to have exploration.
    fixed_parameters : dict of (str, array_like) or collection of, optional
        A dict (or collection of dict, in case of the ``mpc`` wrapping an underlying
        :class:`csnlp.multistart.MultistartNlp` instance) whose keys are the names of
        the MPC parameters and the values are their corresponding values. Use this to
        specify fixed parameters, that is, non-learnable. If ``None``, then no fixed
        parameter is assumed.
    experience : int or ExperienceReplay, optional
        The container for experience replay memory. If ``None`` is passed, then a memory
        with unitary length is created, i.e., it keeps only the latest memory
        transition. If an integer ``n`` is passed, then a memory with the length ``n``
        is created and with sample size ``n``. Otherwise, pass an instance of
        :class:`core.experience.ExperienceReplay` to specify the requirements in more
        details.
    warmstart : "last" or "last-successful" or WarmStartStrategy, optional
        The warmstart strategy for the MPC's NLP. If ``"last-successful"``, the last
        successful solution is used to warmstart the solver for the next iteration.
        If ``"last"``, the last solution is used, regardless of success or failure.
        Furthermore, an instance of :class:`core.warmstart.WarmStartStrategy` can
        be passed to specify a strategy for generating multiple warmstart points for the
        MPC's NLP instance. This is useful to generate multiple initial conditions for
        highly non-convex, nonlinear problems. This feature can only be used with an
        MPC that has an underlying multistart NLP problem (see :mod:`csnlp.multistart`).
    hessian_type : {"none", "natural"}, optional
        The type of hessian to use in this (potentially) second-order algorithm.
        If ``"none"``, no second order information is used. If ``"natural"``, the Fisher
        information matrix is used to perform a natural policy gradient update. This
        option must be in accordance with the choice of ``optimizer``, that is, if the
        optimizer does not use second order information, then this option must be set to
        ``none``.
    rollout_length : int, optional
        Number of steps of each closed-loop simulation, which defines a complete
        trajectory of the states (i.e., a rollout), and is saved in the experience as a
        single item (since LSTD DPG needs to draw samples of trajectories). In case the
        env is episodic, it can be ``-1``, in which case the rollout length coincides
        with the episode's length. In case the env is not episodic, i.e., it never
        terminates, a length ``>0`` must be given in order to know when to save the
        current trajectory as an atomic item in memory.
    record_policy_performance: bool, optional
        If ``True``, the performance of each rollout is stored in the field
        :attr:`policy_performances`, which otherwise is ``None``. By default, does not
        record them.
    record_policy_gradient: bool, optional
        If ``True``, the (estimated) policy gradient of each update is stored in the
        field :attr:`policy_gradients`, which otherwise is `None`. By default, does not
        record them.
    state_features : casadi.Function, optional
        The state feature vector to be used in the linear approximation of the
        value function, which takes the form of

        .. math:: V_v(s) = \Phi(s)^\top v,

        where :math:`s` is the state, :math:`v` are the weights, and :math:`\Phi(s)` is
        the state feature vector. This function is assumed to have one input and one
        output. By default, if not provided, it is designed as all monomials of the
        state with degrees ``<= 2`` (see :func:`util.math.monomials_basis_function`).
    linsolver : "csparse" or "mldivide", optional
        The type of linear solver to be used for solving the linear system derived
        from the KKT conditions and used to estimate the gradient of the policy. By
        default, ``"csparse"`` is chosen as the KKT matrix is most often sparse.
    ridge_regression_regularization : float, optional
        Ridge regression regularization used during the computations of the LSTD
        weights via least-squares. By default, ``1e-6``.
    use_last_action_on_fail : bool, optional
        In case the MPC solver fails
         - if ``False``, the action from the last solver's iteration is returned anyway
           (though suboptimal)
         - if ``True``, the action from the last successful call to the MPC is returned
           instead (if the MPC has been solved at least once successfully).

        By default, ``False``.
    name : str, optional
        Name of the agent. If ``None``, one is automatically created from a counter of
        the class' instancies.

    Raises
    ------
    ValueError
        If the exploration strategy is ``None`` or an instance of ``NoExploration``, as
        DPG requires exploration.
    """

    def __init__(
        self,
        mpc: Mpc[SymType],
        update_strategy: Union[int, UpdateStrategy],
        discount_factor: float,
        optimizer: GradientBasedOptimizer,
        learnable_parameters: LearnableParametersDict[SymType],
        exploration: ExplorationStrategy,
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        hessian_type: Literal["none", "natural"] = "none",
        rollout_length: int = -1,
        record_policy_performance: bool = False,
        record_policy_gradient: bool = False,
        state_features: Optional[cs.Function] = None,
        linsolver: Literal["csparse", "mldivide"] = "csparse",
        ridge_regression_regularization: float = 1e-6,
        use_last_action_on_fail: bool = False,
        name: Optional[str] = None,
    ) -> None:
        if exploration is None or isinstance(exploration, NoExploration):
            raise ValueError("DPG requires exploration, but none was provided.")
        # change default update hook to 'on_episode_end'
        if not isinstance(update_strategy, UpdateStrategy):
            update_strategy = UpdateStrategy(update_strategy, "on_episode_end")
        super().__init__(
            mpc=mpc,
            update_strategy=update_strategy,
            discount_factor=discount_factor,
            optimizer=optimizer,
            learnable_parameters=learnable_parameters,
            fixed_parameters=fixed_parameters,
            exploration=exploration,
            experience=experience,
            warmstart=warmstart,
            use_last_action_on_fail=use_last_action_on_fail,
            name=name,
        )
        self.hessian_type = hessian_type
        self._sensitivity = self._init_sensitivity(linsolver)
        self._Phi = (
            monomials_basis_function(mpc.ns, 0, 2)
            if state_features is None
            else state_features
        )
        self.regularization = ridge_regression_regularization
        self.rollout_length = rollout_length
        self._rollout: list[
            tuple[
                ObsType,
                ActType,
                SupportsFloat,
                ObsType,
                npt.NDArray[np.floating],
            ]
        ] = nb.typed.List()
        self.policy_performances: Optional[list[float]] = (
            [] if record_policy_performance else None
        )
        self.policy_gradients: Optional[list[npt.NDArray[np.floating]]] = (
            [] if record_policy_gradient else None
        )

    def update(self) -> Optional[str]:
        sample = self.experience.sample()
        compute_fisher_mat = self.hessian_type == "natural"
        dJdtheta, fisher_hessian = _estimate_gradient_update(
            sample, self.discount_factor, self.regularization, compute_fisher_mat
        )
        if self.policy_gradients is not None:
            self.policy_gradients.append(dJdtheta)
        return self.optimizer.update(dJdtheta, fisher_hessian)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        truncated = terminated = False
        timestep = 0
        rewards = 0.0
        state = init_state
        rollout_length = self.rollout_length
        action_space = getattr(env, "action_space", None)
        gradient_based_exploration = self.exploration.mode == "gradient-based"
        action_keys = self.V.actions.keys()

        while not (truncated or terminated):
            # compute V(s)
            action, sol = self.state_value(state, False, action_space=action_space)
            if gradient_based_exploration:
                # NOTE: if the exploration affects the gradient, unfortunately we have
                # to solve again the NLP (but deterministically this time)
                action_opt, sol_opt = self.state_value(
                    state, True, action_space=action_space
                )
            else:
                # otherwise, just retrieve manually the unperturbed optimal action
                action_opt = cs.vertcat(*(sol.vals[u][:, 0] for u in action_keys))
                sol_opt = sol

            # step the system with the action just computed
            state_new, cost, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            # store transition in experience
            if sol.success and sol_opt.success:
                # NOTE: according to Cai et al. [3], the sensitivities should naively be
                # computed with the solution of unpertubed MPC (i.e., sol_opt).
                # According to Gros and Zanon [2], it is hinted that the perturbed
                # solution should be used instead (sol).
                exploration = np.asarray((action - action_opt).elements())
                sol_vals = np.asarray(sol_opt.x_and_lam_and_p.elements())
                self._rollout.append((state, exploration, cost, state_new, sol_vals))
            else:
                status = f"{sol.status}/{sol_opt.status}"
                self.on_mpc_failure(episode, timestep, status, raises)

            # increase counters
            state = state_new
            rewards += float(cost)
            timestep += 1

            # first, check if current rollout has reached its length, and only then
            # invoke on_timestep_end (as it might trigger an update)
            if rollout_length > 0 and len(self._rollout) >= rollout_length:
                self._consolidate_rollout_into_memory()
            self.on_timestep_end(env, episode, timestep)

        # consolidate rollout at the end of episode, if no length was specified
        if rollout_length <= 0:
            self._consolidate_rollout_into_memory()
        return rewards

    def _init_sensitivity(self, linsolver: str) -> Callable[[cs.DM, int], np.ndarray]:
        """Internal utility to compute the derivatives w.r.t. the learnable parameters
        and other functions in order to estimate the policy gradient."""
        assert (self.hessian_type == "none" and self.optimizer._order == 1) or (
            self.hessian_type == "natural" and self.optimizer._order == 2
        ), "expected 1st-order (2nd-order) optimizer with `none` (`natural`) hessian"
        nlp = self._V.nlp
        y = nlp.primal_dual
        theta = cs.vvcat(self._learnable_pars.sym.values())
        u0 = cs.vcat(self._V.first_actions.values())
        x_lam_p = cs.vertcat(nlp.primal_dual, nlp.p)

        # compute first bunch of derivatives
        snlp = NlpSensitivity(nlp, theta)
        Kt = snlp.jacobian("K-p")
        Ky = snlp.jacobian("K-y")
        dydu0 = cs.jacobian_sparsity(u0, y).T

        # instantiate linear solver (must be MX, so SX has to be converted)
        if nlp.sym_type is cs.SX:
            x_lam_p, x_lam_p_sx = cs.MX.sym("in", *x_lam_p.shape), x_lam_p
            Kt, Ky = cs.Function("sx2mx", (x_lam_p_sx,), (Kt, Ky))(x_lam_p)
        solver = (
            cs.solve  # cs.mldivide
            if linsolver == "mldivide"
            else cs.Linsol("linsolver", linsolver, Ky.sparsity()).solve
        )

        # compute sensitivity and convert to function (faster runtime)
        dpidtheta = -solver(Ky, Kt).T @ dydu0
        sensitivity = cs.Function(
            "dpidtheta",
            (x_lam_p,),
            (dpidtheta,),
            ("x_lam_p",),
            ("dpidtheta",),
            {"cse": True},
        )
        ntheta, na = dpidtheta.shape

        # wrap to conveniently return arrays. Casadi does not support tensors with
        # >2 dims, so dpidtheta gets squished in the 3rd dim and needs reshaping
        def func(sol_values: cs.DM, N: int) -> np.ndarray:
            return (
                np.ascontiguousarray(sensitivity(sol_values.T).elements())
                .reshape(ntheta, na, N, order="F")
                .transpose((2, 0, 1))
            )

        return func

    def _consolidate_rollout_into_memory(self) -> None:
        """Internal utility to compact current rollout into a single item in memory."""
        # convert rollout to arrays and clear it
        N, S, E, L, vals = _consolidate_rollout(self._rollout, self._V.ns, self._V.na)
        self._rollout.clear()

        # compute Phi, dpidtheta, Psi, and CAFA weight v
        Phi = np.ascontiguousarray(self._Phi(S.T).elements()).reshape(N + 1, -1)
        dpidtheta = self._sensitivity(vals, N)
        Psi = (dpidtheta @ E).reshape(N, dpidtheta.shape[1])
        v = _compute_cafa_weight_v(Phi, L, self.discount_factor, self.regularization)

        # save to experience
        self.store_experience((L, Phi, Psi, dpidtheta, v))
        if self.policy_performances is not None:
            self.policy_performances.append(L.sum())


@nb.njit(cache=True, nogil=True, parallel=True)
def _consolidate_rollout(
    rollout: list[tuple[ObsType, ActType, float, ObsType, np.ndarray]],
    ns: int,
    na: int,
) -> tuple[int, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Internal utility to convert a rollout list to arrays."""
    N = len(rollout)
    S = np.empty((N + 1, ns))
    E = np.empty((N, na, 1))  # additional dim required for Psi
    L = np.empty(N)
    sol_vals = np.empty((N, rollout[0][-1].size))
    for i in nb.prange(N):
        s, e, cost, _, sol_val = rollout[i]
        S[i] = s.reshape(-1)
        E[i, :, 0] = e
        L[i] = cost
        sol_vals[i] = sol_val
    S[-1] = rollout[-1][3].reshape(-1)
    return N, S, E, L, sol_vals


@nb.njit(cache=True, nogil=True)
def _compute_cafa_weight_v(
    Phi_all: np.ndarray, L: np.ndarray, discount_factor: float, regularization: float
) -> np.ndarray:
    """Compute the CAFA weight ``v`` via ridge regression."""
    # solve for v via ridge regression: -phi'(gamma phi+ - phi) v = phi'L
    Phi = Phi_all[:-1]
    Phi_next = Phi_all[1:]
    Phi_diff = discount_factor * Phi_next - Phi
    M = Phi_diff.T @ Phi @ Phi.T
    R = regularization * np.eye(M.shape[0])
    return np.linalg.solve(M @ Phi_diff + R, -M @ L)


@nb.njit(cache=True, nogil=True)
def _compute_cafa_weight_w(
    Phi_all: np.ndarray,
    Psi: np.ndarray,
    L: np.ndarray,
    v: np.ndarray,
    discount_factor: float,
    regularization: float,
) -> np.ndarray:
    """Compute the CAFA weight ``w`` via ridge regression."""
    # solve for w via ridge regression: psi' psi w = psi' (L + (gamma phi+ - phi) v)
    Phi = Phi_all[:-1]
    Phi_next = Phi_all[1:]
    Phi_diff = discount_factor * Phi_next - Phi
    A = Psi.T @ Psi
    b = Psi.T @ (L + Phi_diff @ v)
    R = regularization * np.eye(A.shape[0])
    return np.linalg.solve(A.T @ A + R, A.T @ b)


def _estimate_gradient_update(
    sample: Iterator[ExpType],
    discount_factor: float,
    regularization: float,
    return_fisher_hessian: bool,
) -> tuple[np.ndarray, Optional[np.ndarray]]:
    """Internal utility to estimate the gradient of the policy and possibly the Fisher
    information matrix as well."""
    # compute average v and w
    sample_ = list(sample)  # load whole iterator into a list
    v = np.mean([o[4] for o in sample_], 0)
    w_ = [
        _compute_cafa_weight_w(Phi, Psi, L, v, discount_factor, regularization)
        for L, Phi, Psi, _, _ in sample_
    ]
    w = np.mean(w_, 0)

    if return_fisher_hessian:
        # compute both policy gradient and Fisher information matrix
        fisher_hess_ = []
        dJdtheta_ = []
        for _, _, _, dpidtheta, _ in sample_:
            F = (dpidtheta @ dpidtheta.transpose((0, 2, 1))).sum(0)
            fisher_hess_.append(F)
            dJdtheta_.append(F @ w)
        return np.mean(dJdtheta_, 0), np.mean(fisher_hess_, 0)

    # compute only policy gradient estimate
    dJdtheta_ = [(o[3] @ o[3].transpose((0, 2, 1))).sum(0) @ w for o in sample_]
    return np.mean(dJdtheta_, 0), None



from typing import Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.parameters import LearnableParametersDict, SymType
from ..core.schedulers import Scheduler
from .gradient_based_optimizer import GradientBasedOptimizer, LrType


class RMSprop(GradientBasedOptimizer[SymType, LrType]):
    r"""RMSprop optimizer, based on :cite:`hinton_neural_2012` and
    :class:`torch.optim.RMSprop`.

    While there is an abundance of resources to understand the RMSprop optimizer, how do
    we address the constraints on the learnable parameters? Similarly to
    :class:`mpcrl.optim.GradientDescent`, once the update rule has calculated the change
    :math:`g` in the parameters (in an unconstrained fashion), we force the constraints
    via the Quadratic Programming (QP) problem

    .. math::

        \begin{aligned}
            \min_{\Delta\theta} & \quad \frac{1}{2} \lVert \Delta\theta \rVert_2^2 + \alpha g^\top \Delta\theta \\
            \text{s.t.} & \quad \theta_{\text{lower}} \leq \theta + \Delta\theta \leq \theta_{\text{upper}}
        \end{aligned}

    Parameters
    ----------
    learning_rate : float or array or :class:`mpcrl.core.schedulers.Scheduler`
        The learning rate of the optimizer. It can be:

        - a float, in case the learning rate must stay constant and is the same for all
          learnable parameters

        - an array, in case the learning rate must stay constant but is different for
          each parameter (should have the same size as the number of learnable
          parameters)

        - a :class:`mpcrl.core.schedulers.Scheduler`, in case the learning rate can vary
          during the learning process (usually, it is set to decay). See the ``hook``
          argument for more details on when this scheduler is stepped.
    alpha : float, optional
        A positive float that specifies the decay rate of the running average of the
        gradient. By default, it is set to ``0.99``.
    eps : float, optional
        Term added to the denominator to improve numerical stability. By default, it
        is set to ``1e-8``.
    weight_decay : float, optional
        A positive float that specifies the decay of the learnable parameters in the
        form of an L2 regularization term. By default, it is set to ``0.0``, so no
        decay/regularization takes place.
    momentum : float, optional
        A positive float that specifies the momentum factor. By default, it is set
        to ``0.0``, so no momentum is used.
    centered : bool, optional
        If ``True``, compute the centered RMSProp, i.e., the gradient is normalized by
        an estimation of its variance.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies when to step the optimizer's learning rate's scheduler to decay
        its value. This allows to vary the rate over the learning iterations. The
        options are:

        - ``"on_update"`` steps the learning rate after each agent's update

        - ``"on_episode_end"`` steps the learning rate after each episode's end

        - ``"on_timestep_end"`` steps the learning rate after each env's timestep.

        By default, ``"on_update"`` is selected.
    max_percentage_update : float, optional
        A positive float that specifies the maximum percentage change the learnable
        parameters can experience in each update. For example,
        ``max_percentage_update=0.5`` means that the parameters can be updated by up
        to 50% of their current value. By default, it is set to ``+inf``.
        If specified, the update becomes constrained and has to be solved as a QP, which
        is inevitably slower than its unconstrained counterpart (a linear system).
    bound_consistency : bool, optional
        A boolean that, if ``True``, forces the learnable parameters to lie in their
        bounds when updated. This is done via :func:`numpy.clip`. Only beneficial if
        numerical issues arise during updates, e.g., due to the QP solver not being able
        to guarantee bounds.
    """

    _order = 1
    _hessian_sparsity = "diag"
    # In RMSprop, hessian is at most diagonal, i.e., in case we have constraints

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        alpha: float = 0.99,
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        momentum: float = 0.0,
        centered: bool = False,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
        bound_consistency: bool = False,
    ) -> None:
        super().__init__(learning_rate, hook, max_percentage_update, bound_consistency)
        self.weight_decay = weight_decay
        self.alpha = alpha
        self.eps = eps
        self.momentum = momentum
        self.centered = centered

    def set_learnable_parameters(self, pars: LearnableParametersDict[SymType]) -> None:
        super().set_learnable_parameters(pars)
        # initialize also running averages
        n = pars.size
        self._square_avg = np.zeros(n, dtype=float)
        self._grad_avg = np.zeros(n, dtype=float) if self.centered else None
        self._momentum_buf = np.zeros(n, dtype=float) if self.momentum > 0.0 else None

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value

        # compute candidate update
        weight_decay = self.weight_decay
        lr = self.lr_scheduler.value
        if weight_decay > 0.0:
            gradient = gradient + weight_decay * theta
        dtheta, self._square_avg, self._grad_avg, self._momentum_buf = _rmsprop(
            gradient,
            self._square_avg,
            lr,
            self.alpha,
            self.eps,
            self.centered,
            self._grad_avg,
            self.momentum,
            self._momentum_buf,
        )

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            return theta + dtheta, None
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=cs.DM.eye(theta.shape[0]), g=-dtheta, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _rmsprop(
    grad: np.ndarray,
    square_avg: np.ndarray,
    lr: LrType,
    alpha: float,
    eps: float,
    centered: bool,
    grad_avg: Optional[np.ndarray],
    momentum: float,
    momentum_buffer: Optional[np.ndarray],
) -> tuple[np.ndarray, np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
    """Computes the update's change according to Adam algorithm."""
    square_avg = alpha * square_avg + (1 - alpha) * np.square(grad)

    if centered:
        grad_avg = alpha * grad_avg + (1 - alpha) * grad
        avg = np.sqrt(square_avg - np.square(grad_avg))
    else:
        avg = np.sqrt(square_avg)
    avg += eps

    if momentum > 0.0:
        momentum_buffer = momentum * momentum_buffer + grad / avg
        dtheta = -lr * momentum_buffer
    else:
        dtheta = -lr * grad / avg
    return dtheta, square_avg, grad_avg, momentum_buffer



"""Wrappers for learning and non-learning agents.

First of all, this submodule introduces two base classes for wrappers to be appplied to
agents, not environments. These classes are :class:`mpcrl.wrappers.agents.Wrapper` and
:class:`mpcrl.wrappers.agents.LearningWrapper`. The former is a wrapper for non-learning
agents, while the latter is a wrapper for learning agents. Both classes define an
interface equal to the agent's, so that they can be used as agents themselves
seamlessly.

Outside of these base classes, this submodule provides a few concrete wrappers for
logging information about the agent's learning/evaluation process, recording the updates
of the MPC parametrization during learning, and launching periodic evaluations of the
agent during training.
"""

__all__ = ["Evaluate", "LearningWrapper", "Log", "RecordUpdates", "Wrapper"]

from .evaluate import Evaluate
from .log import Log
from .record_updates import RecordUpdates
from .wrapper import LearningWrapper, Wrapper



"""Error and warning classes for signaling failure in the MPC solver and in the
updates, as well as convenience functions for raising these errors and warnings."""

from warnings import warn as _warn


class MpcSolverError(RuntimeError):
    """Exception class for raising errors when an MPC solver fails."""


class MpcSolverWarning(RuntimeWarning):
    """Warning class for raising errors when an MPC solver fails."""


class UpdateError(RuntimeError):
    """Exception class for raising errors when a learning agent's updates fails."""


class UpdateWarning(RuntimeWarning):
    """Warning class for raising errors when a learning agent's updates fails."""


def raise_or_warn_on_mpc_failure(msg: str, raises: bool) -> None:
    """Raises an error or warning with the given message due to an MPC solver's failure.

    Parameters
    ----------
    msg : str
        The exception or warning message.
    raises : bool
        If ``True``, raises an exception; otherwise, throws a warning.

    Raises
    ------
    UpdateError
        Raises :class:`MpcSolverError` if ``raises=True``; otherwise raises
        :class:`MpcSolverWarning`.
    """
    if raises:
        raise MpcSolverError(msg)
    else:
        _warn(msg, MpcSolverWarning)


def raise_or_warn_on_update_failure(msg: str, raises: bool) -> None:
    """Raises an error or warning with the given message due to an update failure.

    Parameters
    ----------
    msg : str
        The exception or warning message.
    raises : bool
        If ``True``, raises an exception; otherwise, throws a warning.

    Raises
    ------
    UpdateError
        Raises :class:`UpdateError` if ``raises=True``; otherwise raises
        :class:`UpdateWarning`.
    """
    if raises:
        raise UpdateError(msg)
    else:
        _warn(msg, UpdateWarning)



"""A submodule with utility functions for geometry operations."""

from typing import Optional, Union

import numpy as np
import numpy.typing as npt
from scipy.spatial import ConvexHull as _ConvexHull
from scipy.spatial import Delaunay as _Delaunay

from .seeding import RngType


class ConvexPolytopeUniformSampler:
    """Draws samples uniformly at random from a convex polytopic region (samples from
    its interior and surface are possible to obtain).

    Under the hood, the convex hull and triangulation of the given polytope are
    performed (see :class:`scipy.spatial.ConvexHull` and
    :class:`scipy.spatial.Delaunay`). To sample uniformly from the interior, for each
    sample a triangulation simplex is selected with probability proportional to its
    volume, and a point is drawn uniformly from that simplex via a Dirichlet
    distribution. Likewise, to sample from its surface, a facet is instead selected at
    random with probability proportional to its surface.

    Parameters
    ----------
    vertices : array-like of shape (n, d)
        The vertices of the polytope, where ``n`` is the number of vertices and ``d``
        is the dimension of the space.
    incremental : bool, optional
        Allows adding new points incrementally to the convex hull and triangulation.
        This takes up some additional resources.
    disable_triangulation : bool, optional
        Whether to disable the triangulation. This is useful when only surface samples
        are desired. By default, ``False``.
    disable_convex_hull : bool, optional
        Whether to disable the convex hull computation. This is useful when only
        interior samples are desired. By default, ``False``.
    seed : int, sequence of int, seed or :class:`numpy.random.Generator`, optional
        The seed or random number generator to use for sampling.
    """

    def __init__(
        self,
        vertices: npt.ArrayLike,
        incremental: bool = False,
        disable_triangulation: bool = False,
        disable_convex_hull: bool = False,
        seed: Optional[np.random.Generator] = None,
    ) -> None:
        self.seed(seed)
        self._tri_enabled = not disable_triangulation
        self._qhull_enabled = not disable_convex_hull
        if self._qhull_enabled:
            self._qhull = _ConvexHull(vertices, incremental=incremental)
            self._compute_facet_areas_ratios()
        if self._tri_enabled:
            self._triangulation = _Delaunay(vertices, incremental=incremental)
            self._compute_simplex_volumes_ratios()

    def seed(self, seed: RngType = None) -> npt.NDArray[np.floating]:
        """Resets the RNG engine.

        Parameters
        ----------
        seeed : int, sequence of int, seed or :class:`numpy.random.Generator`, optional
            The seed or random number generator to use for sampling.

        Returns
        -------
        array of shape (size1, size2, ..., d)
            The samples drawn from the simplex.

        Raises
        ------
        ValueError
            Raised if the vertices are not ``d+1`` in number.
        """
        self._np_random = np.random.default_rng(seed)

    def add_points(self, points: npt.ArrayLike, restart: bool = False) -> None:
        """Processes a set of additional new points. See also
        :meth:`scipy.spatial.Delaunay.add_points`.

        Parameters
        ----------
        vertices : array-like  of shape (n, d)
            New points to add, where ``n`` is the number of new points and ``d`` is the
            dimension of the space.
        restart : bool, optional
            Whether to restart processing from scratch, rather than adding points
            incrementally.
        """
        if self._qhull_enabled:
            self._qhull.add_points(points, restart)
            self._compute_facet_areas_ratios()
        if self._tri_enabled:
            self._triangulation.add_points(points, restart)
            self._compute_simplex_volumes_ratios()

    def close(self) -> None:
        """Finishes the incremental processing."""
        if self._qhull_enabled:
            self._qhull.close()
        if self._tri_enabled:
            self._triangulation.close()

    def sample_from_interior(
        self, size: Union[int, tuple[int, ...]] = ()
    ) -> npt.NDArray[np.floating]:
        """Sample uniformly from the interior of the polytope defined by the given
        vertices.

        Parameters
        ----------
        size : int or tuple of ints, optional
            The size of the sample array to draw. By default, one sample is drawn.

        Returns
        -------
        array of shape (size1, size2, ..., d)
            The samples drawn from the polytope.

        Raises
        ------
        RuntimeError
            Raised if the triangulation is disabled.
        """
        if not self._tri_enabled:
            raise RuntimeError("Triangulation is disabled.")
        simplices = self._triangulation.points[self._triangulation.simplices]
        n_simplices, d_plus_1 = simplices.shape[:2]
        simplex_idx = self._np_random.choice(n_simplices, size, p=self._vol_ratios)
        selected_simplices = simplices[simplex_idx]
        weights = self._np_random.dirichlet(np.ones(d_plus_1), size)
        return np.matmul(weights[..., None, :], selected_simplices).squeeze(-2)

    def sample_from_surface(
        self, size: Union[int, tuple[int, ...]] = ()
    ) -> npt.NDArray[np.floating]:
        """Sample uniformly from the surface of the polytope defined by the given
        vertices.

        Parameters
        ----------
        size : int or tuple of ints, optional
            The size of the sample array to draw. By default, one sample is drawn.

        Returns
        -------
        array of shape (size1, size2, ..., d)
            The samples drawn from the polytope.

        Raises
        ------
        RuntimeError
            Raised if the convex hull is disabled.
        """
        if not self._qhull_enabled:
            raise RuntimeError("Convex hull is disabled.")
        facets = self._qhull.points[self._qhull.simplices]
        n_facets, d = facets.shape[:2]
        facet_idx = self._np_random.choice(n_facets, size, p=self._area_ratios)
        selected_facets = facets[facet_idx]
        weights = self._np_random.dirichlet(np.ones(d), size)
        return np.matmul(weights[..., None, :], selected_facets).squeeze(-2)

    def _compute_simplex_volumes_ratios(self) -> None:
        """Computes the ratios of the volumes of the simplices in the triangulation."""
        simplices = self._triangulation.points[self._triangulation.simplices]
        if simplices.shape[0] == 1:
            self._vol_ratios = np.ones(1, dtype=float)
        else:
            vol = np.abs(np.linalg.det(simplices[:, 1:] - simplices[:, :1]))
            self._vol_ratios = vol / vol.sum()

    def _compute_facet_areas_ratios(self) -> None:
        """Computes the ratios of the areas of the facets in the convex hull."""
        facets = self._qhull.points[self._qhull.simplices]
        if facets.shape[0] == 1:
            self._area_ratios = np.ones(1, dtype=float)
        else:
            failed = False
            diff = facets[:, 1:] - facets[:, :1]
            gram = diff @ diff.transpose(0, 2, 1)
            try:
                vol = np.sqrt(np.linalg.det(gram))
                area_ratios = vol / vol.sum()
            except np.linalg.LinAlgError:
                failed = True

            if failed or not np.isfinite(area_ratios).all():
                # alternative method to compute area by projecting and adjusting for the
                # projection angle - https://math.stackexchange.com/a/2098632
                # 1) project facets to d-1 simplex in the last axis
                simplices = facets[..., :-1]
                proj_vol = np.abs(np.linalg.det(simplices[:, 1:] - simplices[:, :1]))
                # 2) compute the projected normals
                normals = self._qhull.equations[:, :-1]  # already normalized
                proj_normals = normals[:, :-1]
                # 3) adjust the projected volume for the angle of the normals
                sine = np.linalg.norm(proj_normals, axis=1)
                vol = proj_vol / np.cos(np.arcsin(sine))
                area_ratios = vol / vol.sum()

            self._area_ratios = area_ratios



import unittest
from itertools import product
from unittest.mock import Mock

import casadi as cs
import numpy as np
import torch
from parameterized import parameterized

from mpcrl import LearnableParameter, LearnableParametersDict
from mpcrl import optim as O
from mpcrl import schedulers as S
from mpcrl.optim.gradient_based_optimizer import GradientBasedOptimizer

np.random.seed(10)
torch.manual_seed(0)
torch.use_deterministic_algorithms(True)

SMALL = 1e-270
N_PARAMS = 5
pars = np.random.randn(N_PARAMS)
LEARNABLE_PARS = LearnableParametersDict(
    [LearnableParameter(f"p{i}", 1, pars[i]) for i in range(N_PARAMS)]
)


class DummyOptimizer(GradientBasedOptimizer):
    _hessian_sparsity = "dense"
    _update_solver = None

    def update(self, *args, **kwargs):
        pass


class TestGradientBasedOptimizer(unittest.TestCase):
    def test_init_and_step(self):
        lr = object()
        perc = object()
        opt = DummyOptimizer(learning_rate=lr, max_percentage_update=perc)
        self.assertIsInstance(opt.lr_scheduler, S.Scheduler)
        self.assertIs(opt.lr_scheduler.value, lr)
        self.assertIsNone(opt.hook)
        self.assertIs(opt.max_percentage_update, perc)
        self.assertIsNone(opt.learnable_parameters)
        self.assertIsNone(opt._update_solver)

    def test_init__with_scheduler(self):
        hook = object()
        learning_rate = S.ExponentialScheduler(0.56, 0.99)
        learning_rate.step = Mock()

        opt = DummyOptimizer(learning_rate, hook=hook)
        opt.step()

        self.assertIsInstance(opt.lr_scheduler, S.Scheduler)
        self.assertEqual(opt.lr_scheduler.value, 0.56)
        self.assertIs(opt.hook, hook)
        learning_rate.step.assert_called_once_with()

    def test_set_learnable_parameters(self):
        opt = DummyOptimizer(0.1)
        opt.set_learnable_parameters(LEARNABLE_PARS)
        self.assertIs(opt.learnable_parameters, LEARNABLE_PARS)
        self.assertIsNone(opt._update_solver)

    def test_set_learnable_parameters__with_solver(self):
        opt = DummyOptimizer(0.1, max_percentage_update=0.5)
        opt.set_learnable_parameters(LEARNABLE_PARS)
        self.assertIs(opt.learnable_parameters, LEARNABLE_PARS)
        self.assertIsInstance(opt._update_solver, cs.Function)

    def test_init__creates_update_solver(self):
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.lb = -np.arange(1, N_PARAMS + 1, dtype=float)
        learnable_pars.ub = np.arange(1, N_PARAMS + 1, dtype=float)
        opt = DummyOptimizer(0.1)
        opt.set_learnable_parameters(learnable_pars)
        self.assertIsNotNone(opt._update_solver)

    def test_get_update_bounds__raises__with_no_learnable_parameters(self):
        opt = DummyOptimizer(learning_rate=0.1)
        msg = "'NoneType' object has no attribute 'lb'"
        with self.assertRaises(AttributeError, msg=msg):
            opt._get_update_bounds(np.random.rand(N_PARAMS))

    def test_get_update_bounds(self):
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        lb = -np.arange(1, N_PARAMS + 1, dtype=float)
        ub = theta = np.arange(1, N_PARAMS + 1, dtype=float)
        learnable_pars.lb = lb
        learnable_pars.ub = ub
        opt = DummyOptimizer(learning_rate=0.1)
        opt.set_learnable_parameters(learnable_pars)
        lb_dtheta, ub_dtheta = opt._get_update_bounds(theta)
        np.testing.assert_array_equal(lb_dtheta, lb - theta)
        np.testing.assert_array_equal(ub_dtheta, 0)

    def test_get_update_bounds__with_maximum_update_percentage(self):
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        lb = theta = -np.arange(1, N_PARAMS + 1, dtype=float)
        ub = np.arange(1, N_PARAMS + 1, dtype=float)
        learnable_pars.lb = lb
        learnable_pars.ub = ub
        opt = DummyOptimizer(learning_rate=0.1, max_percentage_update=0.5)
        opt.set_learnable_parameters(learnable_pars)
        lb_dtheta, ub_dtheta = opt._get_update_bounds(theta)
        np.testing.assert_array_equal(lb_dtheta, 0)
        np.testing.assert_array_equal(ub_dtheta, 0.5 * np.abs(theta))


class TestGradientDescent(unittest.TestCase):
    @parameterized.expand([(False,), (True,)])
    def test_update__unconstrained(self, nesterov: bool):
        opt = O.GradientDescent(
            learning_rate=0.1,
            weight_decay=SMALL,
            momentum=SMALL,
            dampening=SMALL,
            nesterov=nesterov,
        )
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS)
        status = opt.update(g)
        self.assertIsNone(status)
        np.testing.assert_array_almost_equal(learnable_pars.value, theta - 0.1 * g)

    @parameterized.expand([(False,), (True,)])
    def test_update__constrained__with_small_bounds(self, nesterov: bool):
        opt = O.GradientDescent(
            learning_rate=0.1,
            weight_decay=SMALL,
            momentum=SMALL,
            dampening=SMALL,
            nesterov=nesterov,
        )
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        lb = learnable_pars.lb = theta - np.abs(theta) * 5e-2
        ub = learnable_pars.ub = theta + np.abs(theta) * 5e-2
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS) * 100
        status = opt.update(g)
        theta_new = learnable_pars.value
        self.assertIsNone(status)
        self.assertTrue(
            np.where(theta_new >= lb, True, np.isclose(theta_new, lb)).all()
        )
        self.assertTrue(
            np.where(theta_new <= ub, True, np.isclose(theta_new, ub)).all()
        )

    @parameterized.expand([(False,), (True,)])
    def test_update__constrained__with_large_bounds(self, nesterov: bool):
        opt = O.GradientDescent(
            learning_rate=0.1,
            weight_decay=SMALL,
            momentum=SMALL,
            dampening=SMALL,
            nesterov=nesterov,
        )
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        learnable_pars.lb = -np.abs(theta) * 100
        learnable_pars.ub = np.abs(theta) * 100
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS) / 100
        status = opt.update(g)
        self.assertIsNone(status)
        np.testing.assert_array_equal(learnable_pars.value, theta - 0.1 * g)


class TestNetwonMethod(unittest.TestCase):
    @parameterized.expand(product((0.0, SMALL), (False, True)))
    def test_update__unconstrained(self, w: float, cho_before_update: bool):
        opt = O.NetwonMethod(0.1, weight_decay=w, cho_before_update=cho_before_update)
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS)
        H = np.random.randn(N_PARAMS, N_PARAMS)
        H = H @ H.T + np.eye(N_PARAMS)
        status = opt.update(g, H)
        self.assertIsNone(status)
        np.testing.assert_array_almost_equal(
            learnable_pars.value, theta - 0.1 * np.linalg.solve(H, g)
        )

    @parameterized.expand(product((0.0, SMALL), (False, True)))
    def test_update__unconstrained__with_identity_hessian(
        self, w: float, cho_before_update: bool
    ):
        opt = O.NetwonMethod(0.1, weight_decay=w, cho_before_update=cho_before_update)
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS)
        H = np.eye(N_PARAMS)
        status = opt.update(g, H)
        self.assertIsNone(status)
        np.testing.assert_array_almost_equal(learnable_pars.value, theta - 0.1 * g)

    @parameterized.expand(product((0.0, SMALL), (False, True)))
    def test_update__constrained__with_large_bounds_with_identity_hessian(
        self, w: float, cho_before_update: bool
    ):
        opt = O.NetwonMethod(0.1, weight_decay=w, cho_before_update=cho_before_update)
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        learnable_pars.lb = -np.abs(theta) * 100
        learnable_pars.ub = np.abs(theta) * 100
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS) / 100
        H = np.eye(N_PARAMS)
        status = opt.update(g, H)
        self.assertIsNone(status)
        np.testing.assert_array_equal(learnable_pars.value, theta - 0.1 * g)

    @parameterized.expand(product((0.0, SMALL), (False, True)))
    def test_update__constrained__with_small_bounds(
        self, w: float, cho_before_update: bool
    ):
        opt = O.NetwonMethod(0.1, weight_decay=w, cho_before_update=cho_before_update)
        learnable_pars = LEARNABLE_PARS.copy(deep=True)
        learnable_pars.value = theta = np.random.randn(N_PARAMS)
        lb = learnable_pars.lb = theta - np.abs(theta) * 5e-2
        ub = learnable_pars.ub = theta + np.abs(theta) * 5e-2
        opt.set_learnable_parameters(learnable_pars)
        g = np.random.randn(N_PARAMS) * 100
        H = np.random.randn(N_PARAMS, N_PARAMS)
        H = H @ H.T
        status = opt.update(g, H)
        theta_new = learnable_pars.value
        self.assertIsNone(status)
        self.assertTrue(
            np.where(theta_new >= lb, True, np.isclose(theta_new, lb)).all()
        )
        self.assertTrue(
            np.where(theta_new <= ub, True, np.isclose(theta_new, ub)).all()
        )


class TestAdam(unittest.TestCase):
    @parameterized.expand(product((0, 0.01), (False, True), (False, True)))
    def test(self, weight_decay: float, decouple_weight_decay: bool, amsgrad: bool):
        # prepare data
        betas = tuple(np.random.uniform(0.9, 1.0, size=2))
        eps = np.random.uniform(1e-8, 1e-6)
        lr = np.random.uniform(1e-4, 1e-3)

        # prepare torch elements
        x = torch.linspace(-np.pi, np.pi, 2, dtype=torch.float64)
        xx_torch = x.unsqueeze(-1).pow(torch.tensor([1, 2, 3]))
        y_actual_torch = torch.sin(x)
        model_torch = torch.nn.Linear(3, 1, dtype=torch.float64)
        loss_fn_torch = torch.nn.MSELoss(reduction="sum")
        cls = torch.optim.AdamW if decouple_weight_decay else torch.optim.Adam
        optimizer_torch = cls(
            params=model_torch.parameters(),
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )

        # prepare mpcrl elements
        xx_dm = cs.DM(xx_torch.detach().numpy())
        A_sym = cs.MX.sym("A", *model_torch.weight.shape)
        b_sym = cs.MX.sym("b", *model_torch.bias.shape)
        y_pred_sym = xx_dm @ A_sym.T + b_sym
        y_actual_dm = cs.DM(y_actual_torch.detach().clone().numpy())
        loss_sym = cs.sumsqr(y_pred_sym - y_actual_dm)
        p_sym = cs.veccat(A_sym, b_sym)
        dldp_sym = cs.gradient(loss_sym, p_sym)
        model_mpcrl = cs.Function(
            "F", [p_sym], [y_pred_sym, loss_sym, dldp_sym], ["p"], ["y", "l", "dldp"]
        )
        A_mpcrl = model_torch.weight.data.detach().clone().numpy().flatten()
        b_mpcrl = model_torch.bias.data.detach().clone().numpy()
        learnable_pars = LearnableParametersDict(
            [
                LearnableParameter("A", A_mpcrl.size, A_mpcrl),
                LearnableParameter("b", b_mpcrl.size, b_mpcrl),
            ]
        )
        optimizer_mpcrl = O.Adam(
            learning_rate=lr,
            max_percentage_update=1e4,  # test constrained
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            decoupled_weight_decay=decouple_weight_decay,
            amsgrad=amsgrad,
        )
        optimizer_mpcrl.set_learnable_parameters(learnable_pars)

        # run test
        cmp = lambda x, y, msg: np.testing.assert_allclose(
            x, y, rtol=1e-6, atol=1e-6, err_msg=msg
        )
        for i in range(20):
            # torch
            y_pred_torch = model_torch(xx_torch).flatten()
            loss_torch = loss_fn_torch(y_pred_torch, y_actual_torch)
            optimizer_torch.zero_grad()
            loss_torch.backward()
            optimizer_torch.step()
            grad_torch = np.concatenate(
                [
                    model_torch.weight.grad.detach().clone().numpy(),
                    model_torch.bias.grad.detach().clone().numpy(),
                ],
                None,
            )

            # mpcrl
            y_pred_mpcrl, loss_mpcrl, grad_mpcrl = model_mpcrl(
                np.concatenate([A_mpcrl, b_mpcrl], None)
            )
            grad_mpcrl = grad_mpcrl.full().flatten()
            status = optimizer_mpcrl.update(grad_mpcrl)
            p_new = learnable_pars.value
            A_mpcrl, b_mpcrl = np.array_split(p_new, [A_mpcrl.size])

            # check
            self.assertIsNone(status)
            cmp(
                y_pred_mpcrl.full().flatten(),
                y_pred_torch.detach().clone().numpy(),
                f"prediction mismatch at iteration {i}",
            )
            cmp(
                float(loss_mpcrl),
                loss_torch.detach().clone().item(),
                f"loss mismatch at iteration {i}",
            )
            cmp(grad_mpcrl, grad_torch, f"gradient mismatch at iteration {i}")
            cmp(
                A_mpcrl,
                model_torch.weight.detach().clone().numpy().reshape(A_mpcrl.shape),
                f"`A` mismatch at iteration {i}",
            )
            cmp(
                b_mpcrl,
                model_torch.bias.detach().clone().numpy().reshape(b_mpcrl.shape),
                f"`b` mismatch at iteration {i}",
            )


class TestRMSprop(unittest.TestCase):
    @parameterized.expand(product((0, 0.1), (0, 0.9), (False, True)))
    def test(self, weight_decay: float, momentum: float, centered: bool):
        # prepare data
        alpha = np.random.uniform(0.9, 0.99)
        eps = np.random.uniform(1e-8, 1e-6)
        lr = np.random.uniform(1e-4, 1e-3)

        # prepare torch elements
        x = torch.linspace(-np.pi, np.pi, 2, dtype=torch.float64)
        xx_torch = x.unsqueeze(-1).pow(torch.tensor([1, 2, 3]))
        y_actual_torch = torch.sin(x)
        model_torch = torch.nn.Linear(3, 1, dtype=torch.float64)
        loss_fn_torch = torch.nn.MSELoss(reduction="sum")
        optimizer_torch = torch.optim.RMSprop(
            params=model_torch.parameters(),
            lr=lr,
            alpha=alpha,
            eps=eps,
            weight_decay=weight_decay,
            momentum=momentum,
            centered=centered,
        )

        # prepare mpcrl elements
        xx_dm = cs.DM(xx_torch.detach().numpy())
        A_sym = cs.MX.sym("A", *model_torch.weight.shape)
        b_sym = cs.MX.sym("b", *model_torch.bias.shape)
        y_pred_sym = xx_dm @ A_sym.T + b_sym
        y_actual_dm = cs.DM(y_actual_torch.detach().clone().numpy())
        loss_sym = cs.sumsqr(y_pred_sym - y_actual_dm)
        p_sym = cs.veccat(A_sym, b_sym)
        dldp_sym = cs.gradient(loss_sym, p_sym)
        model_mpcrl = cs.Function(
            "F", [p_sym], [y_pred_sym, loss_sym, dldp_sym], ["p"], ["y", "l", "dldp"]
        )
        A_mpcrl = model_torch.weight.data.detach().clone().numpy().flatten()
        b_mpcrl = model_torch.bias.data.detach().clone().numpy()
        learnable_pars = LearnableParametersDict(
            [
                LearnableParameter("A", A_mpcrl.size, A_mpcrl),
                LearnableParameter("b", b_mpcrl.size, b_mpcrl),
            ]
        )
        optimizer_mpcrl = O.RMSprop(
            learning_rate=lr,
            alpha=alpha,
            eps=eps,
            weight_decay=weight_decay,
            momentum=momentum,
            centered=centered,
            max_percentage_update=1e4,  # test constrained
        )
        optimizer_mpcrl.set_learnable_parameters(learnable_pars)

        # run test
        cmp = lambda x, y, msg: np.testing.assert_allclose(
            x, y, rtol=1e-6, atol=1e-6, err_msg=msg
        )
        for i in range(20):
            # torch
            y_pred_torch = model_torch(xx_torch).flatten()
            loss_torch = loss_fn_torch(y_pred_torch, y_actual_torch)
            optimizer_torch.zero_grad()
            loss_torch.backward()
            optimizer_torch.step()
            grad_torch = np.concatenate(
                [
                    model_torch.weight.grad.detach().clone().numpy(),
                    model_torch.bias.grad.detach().clone().numpy(),
                ],
                None,
            )

            # mpcrl
            y_pred_mpcrl, loss_mpcrl, grad_mpcrl = model_mpcrl(
                np.concatenate([A_mpcrl, b_mpcrl], None)
            )
            grad_mpcrl = grad_mpcrl.full().flatten()
            status = optimizer_mpcrl.update(grad_mpcrl)
            p_new = learnable_pars.value
            A_mpcrl, b_mpcrl = np.array_split(p_new, [A_mpcrl.size])

            # check
            self.assertIsNone(status)
            cmp(
                y_pred_mpcrl.full().flatten(),
                y_pred_torch.detach().clone().numpy(),
                f"prediction mismatch at iteration {i}",
            )
            cmp(
                float(loss_mpcrl),
                loss_torch.detach().clone().item(),
                f"loss mismatch at iteration {i}",
            )
            cmp(grad_mpcrl, grad_torch, f"gradient mismatch at iteration {i}")
            cmp(
                A_mpcrl,
                model_torch.weight.detach().clone().numpy().reshape(A_mpcrl.shape),
                f"`A` mismatch at iteration {i}",
            )
            cmp(
                b_mpcrl,
                model_torch.bias.detach().clone().numpy().reshape(b_mpcrl.shape),
                f"`b` mismatch at iteration {i}",
            )


if __name__ == "__main__":
    unittest.main()



import logging
import os
import pickle
import sys
import unittest
from operator import neg
from sys import platform
from warnings import filterwarnings

from csnlp.multistart import RandomStartPoint, RandomStartPoints

for folder in (
    "gradient-based-offpolicy",
    "gradient-based-onpolicy",
    "gradient-free",
    "others",
):
    sys.path.append(os.path.join(os.getcwd(), f"examples/{folder}"))

import casadi as cs
import numpy as np
import torch
from acc_with_iccbf import (
    AccEnv,
    create_clf_cbf_qp,
    create_iccbf_qp,
    simulate_controller,
)
from bayesopt import BoTorchOptimizer, CstrEnv, NoisyFilterObservation, get_cstr_mpc
from dpg import LinearMpc as DpgLinearMpc
from dpg import LtiSystem as DpgLtiSystem
from gymnasium.wrappers import TimeLimit, TransformReward
from parameterized import parameterized
from q_learning import LinearMpc as QLearningLinearMpc
from q_learning import LtiSystem as QLearningLtiSystem
from q_learning_offpolicy import LinearMpc as QLearningOffPolicyLinearMpc
from q_learning_offpolicy import LtiSystem as QLearningOffPolicyLtiSystem
from q_learning_offpolicy import get_rollout_generator
from scipy.io import loadmat

from mpcrl import (
    GlobOptLearningAgent,
    LearnableParameter,
    LearnableParametersDict,
    LstdDpgAgent,
    LstdQLearningAgent,
    UpdateStrategy,
    WarmStartStrategy,
)
from mpcrl import exploration as E
from mpcrl.optim import GradientDescent, NetwonMethod
from mpcrl.util.geometry import ConvexPolytopeUniformSampler
from mpcrl.wrappers.agents import Evaluate, Log, RecordUpdates
from mpcrl.wrappers.envs import MonitorEpisodes

torch.set_default_device("cpu")
torch.set_default_dtype(torch.float64)
filterwarnings("ignore", "Mpc failure", module="mpcrl")

DATA = loadmat(f"tests/data_test_examples_{platform}.mat", squeeze_me=True)


class TestExamples(unittest.TestCase):
    @parameterized.expand([(False,), (True,)])
    def test_q_learning__with_copy_and_pickle(self, use_copy: bool):
        mpc = QLearningLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        env = MonitorEpisodes(TimeLimit(QLearningLtiSystem(), max_episode_steps=100))
        agent = Log(
            RecordUpdates(
                LstdQLearningAgent(
                    update_strategy=1,
                    mpc=mpc,
                    learnable_parameters=learnable_pars,
                    discount_factor=mpc.discount_factor,
                    optimizer=NetwonMethod(
                        learning_rate=5e-2,
                        max_percentage_update=1e3,  # does nothing; allows to test qp
                    ),
                    hessian_type="approx",
                    record_td_errors=True,
                )
            ),
            level=logging.DEBUG,
            log_frequencies={"on_timestep_end": 100},
        )

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        J = agent.train(env=env, episodes=1, seed=69).item()
        agent = pickle.loads(pickle.dumps(agent))

        X = env.observations[0].squeeze().T
        U = env.actions[0].squeeze()
        R = env.rewards[0]
        TD = np.squeeze(agent.td_errors)
        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )

        # from scipy.io import savemat
        # DATA.update({
        #     "ql_J": J,
        #     "ql_X": X,
        #     "ql_U": U,
        #     "ql_R": R,
        #     "ql_TD": TD,
        #     "ql_pars": PARS,
        # })
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["ql_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(X, DATA["ql_X"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(U, DATA["ql_U"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(R, DATA["ql_R"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(TD, DATA["ql_TD"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(PARS, DATA["ql_pars"], rtol=1e0, atol=1e0)

    @parameterized.expand([(False,), (True,)])
    def test_dpg__with_copy_and_pickle(self, use_copy: bool):
        mpc = DpgLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        env = MonitorEpisodes(TimeLimit(DpgLtiSystem(), max_episode_steps=200))
        rollout_length = 50
        agent = Log(
            RecordUpdates(
                LstdDpgAgent(
                    mpc=mpc,
                    learnable_parameters=learnable_pars,
                    discount_factor=mpc.discount_factor,
                    optimizer=GradientDescent(
                        learning_rate=1e-7,
                        max_percentage_update=1e3,  # does nothing; allows to test qp
                    ),
                    update_strategy=UpdateStrategy(rollout_length, "on_timestep_end"),
                    rollout_length=rollout_length,
                    exploration=E.GreedyExploration(0.05),
                    record_policy_performance=True,
                    record_policy_gradient=True,
                )
            ),
            level=logging.DEBUG,
            log_frequencies={"on_timestep_end": 200},
        )

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        J = agent.train(env=env, episodes=1, seed=69).item()
        agent = pickle.loads(pickle.dumps(agent))

        X = env.observations[0].squeeze().T
        U = env.actions[0].squeeze()
        R = env.rewards[0]
        Jest = np.asarray(agent.policy_performances)
        Gest = np.asarray(agent.policy_gradients)
        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )

        # from scipy.io import savemat
        # DATA.update({
        #     "dpg_J": J,
        #     "dpg_X": X,
        #     "dpg_U": U,
        #     "dpg_R": R,
        #     "dpg_Jest": Jest,
        #     "dpg_Gest": Gest,
        #     "dpg_pars": PARS,
        # })
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["dpg_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(X, DATA["dpg_X"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(U, DATA["dpg_U"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(R, DATA["dpg_R"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(Jest, DATA["dpg_Jest"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(Gest, DATA["dpg_Gest"], rtol=1e3, atol=1e3)
        np.testing.assert_allclose(PARS, DATA["dpg_pars"], rtol=1e0, atol=1e0)

    @parameterized.expand([(False,), (True,)])
    def test_bayesopt__with_copy_and_pickle(self, use_copy: bool):
        torch.manual_seed(0)
        env = MonitorEpisodes(TimeLimit(CstrEnv(4e3), max_episode_steps=10))
        env = TransformReward(env, neg)
        env = NoisyFilterObservation(env, [1, 2])
        multistarts = 4
        mpc = get_cstr_mpc(env, horizon=7, multistarts=multistarts, n_jobs=multistarts)
        pars = mpc.parameters
        Y = mpc.variables["y"].shape
        U = mpc.variables["u"].shape
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(n, pars[n].shape, (ub + lb) / 2, lb, ub, pars[n])
                for n, lb, ub in [("narx_weights", -2, 2), ("backoff", 0, 5)]
            )
        )
        warmstart = WarmStartStrategy(
            random_points=RandomStartPoints(
                {
                    "y": RandomStartPoint("normal", scale=[[1.0], [20.0]], size=Y),
                    "u": RandomStartPoint("normal", scale=5.0, size=U),
                },
                biases={
                    "y": CstrEnv.x0[[1, 2]].reshape(-1, 1),
                    "u": sum(CstrEnv.inflow_bound) / 2,
                },
                multistarts=multistarts,
            ),
        )
        agent = GlobOptLearningAgent(
            mpc=mpc,
            learnable_parameters=learnable_pars,
            optimizer=BoTorchOptimizer(initial_random=2, seed=42),
            warmstart=warmstart,
        )
        agent = RecordUpdates(agent)
        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy

        J = agent.train(env=env, episodes=6, seed=69, raises=False)
        agent = pickle.loads(pickle.dumps(agent))

        X = np.squeeze(env.get_wrapper_attr("observations"))
        U = np.squeeze(env.get_wrapper_attr("actions"), (2, 3))
        R = np.squeeze(env.get_wrapper_attr("rewards"))

        # from scipy.io import savemat
        # DATA.update({"bo_J": J, "bo_X": X, "bo_U": U, "bo_R": R})
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["bo_J"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(X, DATA["bo_X"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(U, DATA["bo_U"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(R, DATA["bo_R"], rtol=1e-2, atol=1e-2)

    @parameterized.expand([(False,), (True,)])
    def test_q_learning_offpolicy__with_copy_and_pickle(self, use_copy: bool):
        mpc = QLearningOffPolicyLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        seed = np.random.default_rng(69)
        agent = Evaluate(
            Log(
                RecordUpdates(
                    LstdQLearningAgent(
                        mpc=mpc,
                        learnable_parameters=learnable_pars,
                        discount_factor=mpc.discount_factor,
                        update_strategy=1,
                        optimizer=NetwonMethod(learning_rate=5e-2),
                        hessian_type="approx",
                        record_td_errors=True,
                        remove_bounds_on_initial_action=True,
                    )
                ),
                level=logging.DEBUG,
                log_frequencies={"on_episode_end": 1},
            ),
            eval_env=TimeLimit(QLearningOffPolicyLtiSystem(), 100),
            hook="on_episode_end",
            frequency=2,
            n_eval_episodes=2,
            seed=seed,
        )
        generate_rollout = get_rollout_generator(rollout_seed=69)

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        agent.train_offpolicy(
            episode_rollouts=(generate_rollout(n) for n in range(6)), seed=seed
        )
        J = np.asarray(agent.eval_returns)
        agent = pickle.loads(pickle.dumps(agent))

        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )
        TD = np.squeeze(agent.td_errors)

        # from scipy.io import savemat
        # DATA.update({"ql_offpol_J": J, "ql_offpol_TD": TD, "ql_offpol_pars": PARS})
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["ql_offpol_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(TD, DATA["ql_offpol_TD"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(PARS, DATA["ql_offpol_pars"], rtol=1e0, atol=1e0)

    def test_iccbf(self):
        Tfin = 10
        timesteps = 200
        env = AccEnv(Tfin / timesteps)
        clf_cbf_qp_ctrl = create_clf_cbf_qp(env)
        S1, A1 = simulate_controller(env, clf_cbf_qp_ctrl, timesteps)
        iccbf_qp_ctrl = create_iccbf_qp(env)
        S2, A2 = simulate_controller(env, iccbf_qp_ctrl, timesteps)

        X = np.hstack([S1, S2]).T
        U = np.vstack([A1, A2])
        np.testing.assert_allclose(X, DATA["iccbf_X"])
        np.testing.assert_allclose(U, DATA["iccbf_U"])

    @parameterized.expand([(False,), (True,)])
    def test_polytope_sampling(self, incrm: bool):
        np_random = np.random.default_rng(69)
        ndim = np_random.integers(2, 7)
        nvertices = np_random.integers(10, 15)
        VERTICES = np_random.normal(size=(nvertices, ndim))

        n_samples = tuple(np_random.integers(2, 10, size=2))
        if incrm:
            m = nvertices // 2
            sampler = ConvexPolytopeUniformSampler(VERTICES[:m], incremental=True)
            for i in range(m, nvertices):
                sampler.add_points(VERTICES[i, None])
            sampler.close()
        else:
            sampler = ConvexPolytopeUniformSampler(VERTICES, incremental=False)
        sampler.seed(np_random)
        int_samples = sampler.sample_from_interior(n_samples)
        surf_samples = sampler.sample_from_surface(n_samples)
        samples = np.asarray((int_samples, surf_samples))
        np.testing.assert_allclose(samples, DATA[f"polytope_samples_{int(incrm)}"])


if __name__ == "__main__":
    unittest.main()



from typing import Any, Callable, Generic, Union

from csnlp.util.io import SupportsDeepcopyAndPickle

from ...agents.common.agent import Agent, SymType
from ...agents.common.learning_agent import ExpType, LearningAgent
from ...core.callbacks import CallbackMixin


class Wrapper(SupportsDeepcopyAndPickle, CallbackMixin, Generic[SymType]):
    """Wraps an instance of :class:`mpcrl.Agent` to allow a modular transformation of
    its behaviour. This class is the base class for all wrappers. The subclass could
    override some methods to change the behavior of the original agent without touching
    the original code.

    Parameters
    ----------
    agent : Agent or subclass
        The agent to wrap.
    """

    def __init__(self, agent: Agent[SymType]) -> None:
        SupportsDeepcopyAndPickle.__init__(self)
        CallbackMixin.__init__(self)
        del self._hooks  # keep only one dict of hooks, i.e., the agent's one
        self.agent = agent
        self._hooked_callbacks: dict[str, list[str]] = {}
        self._establish_callback_hooks()

    @property
    def unwrapped(self) -> Union[Agent[SymType], LearningAgent[SymType, ExpType]]:
        """Returns the original agent wrapped by this wrapper."""
        return self.agent.unwrapped

    def is_wrapped(self, wrapper_type: type["Wrapper[SymType]"]) -> bool:
        """Gets whether the agent instance is wrapped or not by the wrapper type.

        Parameters
        ----------
        wrapper_type : type of Wrapper
            Type of wrapper to check if the agent is wrapped with.

        Returns
        -------
        bool
            ``True`` if wrapped by an instance of ``wrapper_type``; ``False``,
            otherwise.
        """
        if isinstance(self, wrapper_type):
            return True
        return self.agent.is_wrapped(wrapper_type)

    def _hook_callback(
        self, attachername: str, callbackname: str, func: Callable[..., None]
    ) -> None:
        # store the callback id for later removal via `detach_wrapper(s)`
        self._hooked_callbacks.setdefault(callbackname, []).append(attachername)
        self.unwrapped._hook_callback(attachername, callbackname, func)

    def detach_wrapper(
        self, recursive: bool = False
    ) -> Union[Agent[SymType], LearningAgent[SymType, ExpType], "Wrapper[SymType]"]:
        """Detaches the wrapper from the agent, returning the unwrapped agent. De facto,
        this method detaches all the hooks attached by this wrapper.

        Parameters
        ----------
        recursive : bool, optional
            If ``True``, detaches all the wrappers around the agent recursively.

        Returns
        -------
        Agent or LearningAgent or Wrapper
            Returns the wrapped agent (or other wrapper) instance. This instance has no
            more active hooks attached by this wrapper. If ``recursive=True``, all the
            wrappers around the agent and their hooks are detached.

        Notes
        -----
        Detaching a wrapper is useful when you want to make sure that the wrapper's
        hooked callback cannot modify the behaviour or data of the agent, for example,
        after learning is done and you want to save and evaluate your learnt policy.
        """
        hooks = self.unwrapped._hooks
        hooked_callbacks = self._hooked_callbacks

        # for each callback type, remove the hooks attached by this wrapper
        for callbackname, attachernames in hooked_callbacks.items():
            hook_group = hooks[callbackname]
            for attachername in attachernames:
                hook_group.pop(attachername)

            # if the callback has no more hooks, remove it
            if not hook_group:
                hooks.pop(callbackname)

        # clear hooked callbacks tracking
        hooked_callbacks.clear()
        return (
            self.agent.detach_wrapper(True)
            if recursive
            and hasattr(self.agent, "detach_wrapper")
            and callable(self.agent.detach_wrapper)
            else self.agent
        )

    def __getattr__(self, name: str) -> Any:
        """Reroutes attributes to the wrapped agent instance."""
        if name.startswith("_"):
            raise AttributeError(f"Accessing private attribute '{name}' is prohibited.")
        return getattr(self.agent, name)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}{self.agent.__str__()}>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}{self.agent.__repr__()}>"


class LearningWrapper(Wrapper[SymType], Generic[SymType, ExpType]):
    """A :class:`Wrapper` subclass dedicated to wrapping instances of
    :class:`mpcrl.LearningAgent`."""

    def __init__(self, agent: LearningAgent[SymType, ExpType]) -> None:
        Wrapper.__init__(self, agent)
        self.agent: LearningAgent[SymType, ExpType]

    @property
    def unwrapped(self) -> LearningAgent[SymType, ExpType]:
        return self.agent.unwrapped



from abc import ABC
from typing import Any, Generic, Optional

from gymnasium import Env

from mpcrl.agents.common.agent import ActType, ObsType

from ...core.update import UpdateStrategy
from ...optim.gradient_free_optimizer import GradientFreeOptimizer
from .agent import SymType
from .learning_agent import LearningAgent


class GlobOptLearningAgent(LearningAgent[SymType, None], ABC, Generic[SymType]):
    """Class for learning agents that employ gradient-free Global Optimization
    strategies (e.g., Bayesian Optimization) to learn/improve the MPC policy.

    Contrary to :class:`RlLearningAgent`, this class does not require a discount factor,
    but requires an instance of a :class:`optim.GradientFreeOptimizer` that
    adheres to the ask-tell interface, i.e., it must implement the
    :func:`optim.GradientFreeOptimizer.ask` and
    :func:`optim.GradientFreeOptimizer.tell` methods.

    Parameters
    ----------
    optimizer : GradientFreeOptimizer
        An instance of :class:`optim.GradientFreeOptimizer` optimizer to ask for
        a suggested set of  parameters to try out, and later tell the value of the
        objective function for that suggested set of parameters.
    kwargs
        Additional arguments to be passed to :class:`LearningAgent`.

        Note: the following kwargs are not yet supported
         - ``"experience"``: usually, GO strategies do not require experience replay
         - ``"update_strategy"``: updates are fixed at the end of each episode.
    """

    def __init__(
        self, optimizer: GradientFreeOptimizer[SymType], **kwargs: Any
    ) -> None:
        for key in ("experience", "update_strategy"):
            if key in kwargs:
                raise ValueError(
                    f"{self.__class__.__name__} does not yet support `{key}` kwargs."
                )
        self.optimizer = optimizer
        super().__init__(update_strategy=UpdateStrategy(1, "on_episode_end"), **kwargs)
        self.optimizer.set_learnable_parameters(self._learnable_pars)

    def train(self, *args: Any, **kwargs: Any) -> Any:
        # prime the initial value of the learnable parameters by asking it to the
        # optimizer. This is the usual way to start Global Optimization strategies,
        # whereas in RL we usually start the user's given initial values.
        self.update()
        return super().train(*args, **kwargs)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        # simply evaluate the MPC on the env with the current set of parameters for one
        # episode, and then tell the optimizer the value of the objective function
        rewards = 0.0
        state = init_state
        truncated, terminated, timestep = False, False, 0

        while not (truncated or terminated):
            action, sol = self.state_value(state, False)
            if not sol.success:
                self.on_mpc_failure(episode, timestep, sol.status, raises)

            state, r, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            rewards += float(r)
            timestep += 1
            self.on_timestep_end(env, episode, timestep)

        values = (
            self._learnable_pars.value_as_dict
            if self.optimizer.prefers_dict
            else self._learnable_pars.value
        )
        self.optimizer.tell(values, rewards)
        return rewards

    def update(self) -> Optional[str]:
        # simply ask the optimizer for a new set of parameters, and then update the
        # current parameters with this new set
        theta_new, status = self.optimizer.ask()
        self._learnable_pars.update_values(theta_new)
        return status



"""Given an MPC controller with several symbolic parameters (some meant to be learned,
some other not), we need a way to specify to the agent of choice which of these are
indeed learnable. This is done by the use of the two classes introduced in this
submodule.

Namely, :class:`LearnableParameter` allows to embed a single parameter and its
information, while :class:`LearnableParametersDict` is a dictionary-like class that
contains several of these :class:`LearnableParameter` instances, and offers different
properties and methods to manage them in bulk.

See also :ref:`user_guide_learnable_parameters`."""

from collections.abc import Iterable
from functools import cached_property
from itertools import chain
from numbers import Integral
from typing import Any, Generic, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from csnlp.core.cache import invalidate_cache
from csnlp.util.io import SupportsDeepcopyAndPickle

from ..util.math import summarize_array

SymType = TypeVar("SymType")  # most likely, T is cs.SX or MX


class LearnableParameter(SupportsDeepcopyAndPickle, Generic[SymType]):
    """A parameter that is learnable, that is, it can be adjusted via RL or any other
    learning strategy. This class is useful for managing symbols, bounds and value of
    the learnable parameter.

    Parameters
    ----------
    name : str
        Name of the learnable parameter.
    shape : int or tuple of ints
        Shape of the parameter.
    value : array_like
        Starting value of the parameter.
    lb : array_like, optional
        Lower bound of the parameter values. If not specified, it is unbounded.
    ub : array_like, optional
        Upper bound of the parameter values. If not specified, it is unbounded.
    sym : T, optional
        An optional reference to a symbolic variable representing this parameter.

    Raises
    ------
    ValueError
        Raises if ``value``, ``lb`` or ``ub`` cannot be broadcasted to a 1D vector with
        shape equal to ``shape``; or if the shape of the symbolic variable ``sym`` does
        not match the shape of the parameter.
    """

    def __init__(
        self,
        name: str,
        shape: Union[int, tuple[int, ...]],
        value: npt.ArrayLike,
        lb: npt.ArrayLike = -np.inf,
        ub: npt.ArrayLike = +np.inf,
        sym: Optional[SymType] = None,
    ) -> None:
        super().__init__()
        self.name = name
        self.shape: tuple[int, ...] = (shape,) if isinstance(shape, Integral) else shape
        self.sym = sym
        self.lb: npt.NDArray[np.floating] = np.broadcast_to(lb, shape)
        self.ub: npt.NDArray[np.floating] = np.broadcast_to(ub, shape)
        self._check_sym_shape()
        self._update_value(value)

    @property
    def size(self) -> int:
        """Gets the number of elements in the parameter."""
        return np.prod(self.shape, dtype=int).item()

    def _update_value(self, v: npt.ArrayLike, **is_close_kwargs: Any) -> None:
        """Internal utility for updating the parameter value with a new value.

        Parameters
        ----------
        new_value : array_like
            New value of the parameter.
        is_close_kwargs
            Additional kwargs for :func:`numpy.isclose`, e.g., ``rtol`` and ``atol``,
            for checking numerical values close to a bound.

        Raises
        ------
        ValueError
            Raises if ``new_value`` cannot be broadcasted to a 1D vector with shape
            equal to ``shape``; or if it does not lie inside the upper and lower bounds
            within the specified tolerances.
        """
        v = np.broadcast_to(v, self.shape)
        lb = self.lb
        ub = self.ub
        if ((v < lb) & ~np.isclose(v, lb, **is_close_kwargs)).any() or (
            (v > ub) & ~np.isclose(v, ub, **is_close_kwargs)
        ).any():
            raise ValueError(
                f"Updated parameter `{self.name}` outside bounds: {lb} <= {v} <= {ub}."
            )
        self.value: npt.NDArray[np.floating] = np.clip(v, lb, ub)

    def _check_sym_shape(self) -> None:
        """Internal utility for checking that the shape of the symbolic variable matches
        the shape of the parameter."""
        if self.sym is None or not hasattr(self.sym, "shape"):
            return
        sym_shape = tuple(self.sym.shape)
        shape = self.shape + tuple(1 for _ in range(len(sym_shape) - len(self.shape)))
        if sym_shape != shape:
            raise ValueError("Shape of `sym` does not match `shape`.")

    def __str__(self) -> str:
        return f"<{self.name}(shape={self.shape})>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(name={self.name},shape={self.shape})>"


class LearnableParametersDict(
    dict[str, LearnableParameter[SymType]], SupportsDeepcopyAndPickle
):
    """:class:`dict`-based collection of :class:`LearnableParameter` instances that
    simplifies the process of managing and updating these. The dict contains pairs of
    parameter's name and parameter's instance.

    Parameters can be retrieved as a normal dictionary by their names, but the class
    also offers several properties that are useful for managing the parameters in bulk,
    such as :attr:`lb`, :attr:`ub`, :attr:`value`, :attr:`value_as_dict` and
    :attr:`sym`. With a single call to :meth:`update_values`, the values of the
    parameters can also be updated.

    Parameters
    ----------
    pars : iterable of :class:`LearnableParameter`, optional
        An optional iterable of parameters to insert into the dict by their names.

    Notes
    -----
    To speed up computations, properties of this class are often cached for faster
    calls to the same methods. However, these are automatically cleared when the
    underlying dict is modified.
    """

    def __init__(self, pars: Optional[Iterable[LearnableParameter[SymType]]] = None):
        if pars is None:
            dict.__init__(self)
        else:
            dict.__init__(self, map(lambda p: (p.name, p), chain(pars)))
        SupportsDeepcopyAndPickle.__init__(self)

    @cached_property
    def size(self) -> int:
        """Gets the overall size of all the learnable parameters."""
        return sum(p.size for p in self.values())

    @cached_property
    def lb(self) -> npt.NDArray[np.floating]:
        """Gets the lower bound of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.lb.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def ub(self) -> npt.NDArray[np.floating]:
        """Gets the upper bound of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.ub.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def value(self) -> npt.NDArray[np.floating]:
        """Gets the values of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.value.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def value_as_dict(self) -> dict[str, npt.NDArray[np.floating]]:
        """Gets the values of all the learnable parameters as a :class:`dict`."""
        return {p.name: p.value for p in self.values()}

    @cached_property
    def sym(self) -> dict[str, Optional[SymType]]:
        """Gets symbols of all the learnable parameters, in a dict. If one parameter
        does not possess the symbol, ``None`` is put."""
        return {
            parname: None if par.sym is None else par.sym
            for parname, par in self.items()
        }

    @invalidate_cache(value, value_as_dict)
    def update_values(
        self,
        new_values: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        **is_close_kwargs: Any,
    ) -> None:
        """Updates the value of each parameter

        Parameters
        ----------
        new_values : array_like or dict of (str, array_like)
            The parameters' new values, either as a single concatenated array (which
            will be splitted according to the sizes and each piece sequentially assigned
            to each parameter), or as a dict of parameter's name vs parameter's new
            value.
        is_close_kwargs
            Additional kwargs for :func:`numpy.isclose`, e.g., ``rtol`` and ``atol``,
            for checking numerical values of parameters close to a bound.

        Raises
        ------
        ValueError
            In case of array-like, raises if ``new_values`` cannot be split according to
            the sizes of parameters; or if the new values cannot be broadcasted to 1D
            vectors according to each parameter's size; or if the new values lie outside
            either the lower or upper bounds of each parameter.
        """
        if isinstance(new_values, dict):
            for parname, new_value in new_values.items():
                self[parname]._update_value(new_value, **is_close_kwargs)
        else:
            cumsizes = np.cumsum([p.size for p in self.values()])[:-1]
            values_ = np.split(new_values, cumsizes)
            for par, val in zip(self.values(), values_):
                par._update_value(val.reshape(par.shape, order="F"), **is_close_kwargs)

    __cache_decorator = invalidate_cache(size, lb, ub, value, value_as_dict, sym)

    @__cache_decorator
    def __setitem__(self, name: str, par: LearnableParameter[SymType]) -> None:
        assert name == par.name, f"Key '{name}' must match parameter name '{par.name}'."
        return super().__setitem__(name, par)

    @__cache_decorator
    def update(
        self,
        pars: Iterable[LearnableParameter[SymType]],
        *args: LearnableParameter[SymType],
    ) -> None:
        return super().update(map(lambda p: (p.name, p), chain(pars, args)))

    @__cache_decorator
    def setdefault(
        self,
        par: LearnableParameter[SymType],
    ) -> LearnableParameter[SymType]:
        return super().setdefault(par.name, par)

    __delitem__ = __cache_decorator(dict.__delitem__)
    pop = __cache_decorator(dict.pop)
    popitem = __cache_decorator(dict.popitem)
    clear = __cache_decorator(dict.clear)

    def copy(
        self, deep: bool = False, invalidate_caches: bool = True
    ) -> "LearnableParametersDict[SymType]":
        """Creates a shallow or deep copy of the dict of learnable parameters.

        Parameters
        ----------
        deep : bool, optional
            If ``True``, a deepcopy of the dict and its parameters is returned;
            otherwise, the copy is only shallow.
        invalidate_caches : bool, optional
            If `True`, methods decorated with :func:`csnlp.core.cache.invalidate_cache`
            are called to clear cached properties/lru caches in the copied instance.
            Otherwise, caches in the copy are not invalidated. By default, ``True``.
            Only relevant when ``deep=True``.

        Returns
        -------
        LearnableParametersDict[T]
            A copy of the dict of learnable parameters.
        """
        return (
            SupportsDeepcopyAndPickle.copy(self, invalidate_caches)
            if deep
            else LearnableParametersDict[SymType](self.values())
        )

    def stringify(
        self, summarize: bool = True, precision: int = 3, ddof: int = 0
    ) -> str:
        """Returns a string representing the dict of learnable parameters.

        Parameters
        ----------
        summarize : bool, optional
            If ``True`` (default), array parameters are summarized; otherwise, the
            entire array is printed.
        precision : int, optional
            The printing precision of floating point numbers.
        ddof : int, optional
            Degrees of freedom for computing standard deviations (see
            :func:`numpy.std`).

        Returns
        -------
        str
            A string representing the dict and its parameters.
        """

        def p2s(p: LearnableParameter) -> str:
            if p.size == 1:
                return f"{p.name}={p.value.item():.{precision}f}"
            if summarize:
                return f"{p.name}: {summarize_array(p.value, precision, ddof)}"
            return np.array2string(p.value, precision=precision)

        return "; ".join(p2s(p) for p in self.values())

    def __str__(self) -> str:
        return self.stringify()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {super().__repr__()}>"



r"""A module with utility functions and classes around and for creating control
problems, agents and their components, and typing.

Submodules
==========

.. autosummary::
   :toctree: generated
   :template: module.rst
   :caption: Submodules

   control
   geometry
   iters
   math
   named
   seeding
"""

__all__ = ["control", "geometry", "iters", "math", "named", "seeding"]

from . import control, geometry, iters, math, named, seeding



r"""Exploration is a fundamental concept in Reinforcement Learning. Without it, often
the learning algorithms converge to very suboptimal solutions, or don't even work.

This submodule contains base classes and implementations for exploration strategies in
the context of MPC-based RL. These classes allow the agent to draw perturbations to
apply then to the MPC's optimal action, thus inducing exploration. Mathematically
speaking, this can be achieved in two distinct ways, or modes:

- **additive**: this is the simplest way to apply perturbations. When the MPC solver
  provides the optimal action to take in the environment's current state :math:`s` by
  solving the state value function problem :math:`\min_{u} V(s)`, before applying the
  action to the environment, the agent will draw a perturbation :math:`p` from the
  exploration stratey and apply the action :math:`\tilde{u} = u + p` to the environment.

- **gradient-based**: this is a more sophisticated way to apply perturbations. We can
  induce exploration more safely by modifying the objective of the state value function
  as :math:`\min_{u} V(s) + p^\top u_0`, where :math:`p` is the random perturbation and
  :math:`u_0` is the first action in the NLP problem. This way, we can perturb the
  gradient of the solution based on the scale of the first action.

See :ref:`user_guide_exploration` for a more thorough explanation. In any case,
whichever mode is selected, all the modifications and perturbations are taken care of
automatically by the agent and the exploration strategy."""

from abc import ABC, abstractmethod
from typing import Any, Literal, Optional, Union

import numpy as np
import numpy.typing as npt

from ..util.seeding import RngType
from .schedulers import NoScheduling, Scheduler


class ExplorationStrategy(ABC):
    """Base abstract class for exploration strategies.

    Parameters
    ----------
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook onto, i.e., when to step the exploration's
        schedulers (if any) to, e.g., decay the chances of exploring or the perturbation
        strength (see :meth:`step` also). The options are

        - ``"on_update"``, which steps the exploration after each agent's update
        - ``"on_episode_end"``, which steps the exploration after each episode ends
        - ``"on_timestep_end"``, which steps the exploration after each env's timestep.

        By default, ``"on_update"`` is selected.
    mode : {"gradient-based", "additive"} optional
        Mode of application of explorative perturbations to the MPC. If ``"additive"``,
        then the drawn pertubation is added to the optimal action computed by the
        MPC solver. By default, ``"gradient-based"`` is selected, and in this mode the
        pertubations enter  directly in the MPC objective and is multiplied by the first
        action, thus affecting its gradient.
    """

    def __init__(
        self,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
    ) -> None:
        super().__init__()
        self._hook = hook
        self._mode = mode

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        """Gets which callback the exploration is hooked on, i.e., when to step the
        exploration's schedulers (if any) to, e.g., decay the chances of exploring or
        the perturbation strength (see :meth:`step` also). Can be ``None`` in case no
        hook is needed."""
        return self._hook

    @property
    def mode(self) -> Literal["gradient-based", "additive"]:
        """Gets the mode of application of explorative perturbations to the MPC."""
        return self._mode

    @abstractmethod
    def can_explore(self) -> bool:
        """Computes whether, according to the exploration strategy, the agent should
        explore or not now, at the current instant.

        Returns
        -------
        bool
            ``True`` if the agent should explore according to this strategy; otherwise,
            ``False``.
        """

    @abstractmethod
    def step(self, *args: Any, **kwargs: Any) -> None:
        """Steps (i.e., decays or increases) any scheduler that this class holds, e.g.,
        exploration's strength and chances."""

    @abstractmethod
    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        """Returns a random perturbation."""

    def reset(self, _: RngType = None) -> None:
        """Resets the exploration status, in case it is non-deterministic."""

    def __str__(self) -> str:
        return self.__class__.__name__

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(hook={self.hook},mode={self.mode})"


class NoExploration(ExplorationStrategy):
    """Strategy where no exploration is allowed at any time or, in other words, the
    policy is always deterministic (only based on the current state, and not perturbed).

    Notes
    -----
    This is a special kind of :class:`ExplorationStrategy`, the only one without any
    :attr:`hook` and :attr:`mode`.
    """

    def __init__(self) -> None:
        super().__init__()
        del self._hook, self._mode

    @property
    def hook(self) -> None:
        """Returns ``None``, since no exploration is allowed."""
        return None

    @property
    def mode(self) -> None:
        """Returns ``None``, since no exploration is allowed."""
        return None

    def can_explore(self) -> bool:
        return False

    def step(self, *_, **__) -> None:
        """Does nothing, since no exploration is allowed."""
        return

    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        raise NotImplementedError(
            f"Perturbation not implemented in {self.__class__.__name__}"
        )

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"


class GreedyExploration(ExplorationStrategy):
    """Fully greedy strategy that always perturbs randomly the MPC policy.

    Parameters
    ----------
    strength : scheduler or array/supports-algebraic-operations
        The strength of the exploration. If passed in the form of an
        :class:`mpcrl.schedulers.Scheduler`, then the strength can be scheduled to decay
        or increase every time :meth:`step` is called. Otherwise, it is kept constant.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook onto, i.e., when to step the exploration's
        schedulers (if any) to, e.g., decay the chances of exploring or the perturbation
        strength (see :meth:`step` also). The options are

        - ``"on_update"``, which steps the exploration after each agent's update
        - ``"on_episode_end"``, which steps the exploration after each episode ends
        - ``"on_timestep_end"``, which steps the exploration after each env's timestep.

        By default, ``"on_update"`` is selected.
    mode : {"gradient-based", "additive"} optional
        Mode of application of explorative perturbations to the MPC. If ``"additive"``,
        then the drawn pertubation is added to the optimal action computed by the
        MPC solver. By default, ``"gradient-based"`` is selected, and in this mode the
        pertubations enter  directly in the MPC objective and is multiplied by the first
        action, thus affecting its gradient.
    seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
        Number to seed the :class:`numpy.random.Generator` used for randomizing the
        exploration. By default, ``None``.
    """

    def __init__(
        self,
        strength: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        super().__init__(hook, mode)
        if not isinstance(strength, Scheduler):
            strength = NoScheduling[npt.NDArray[np.floating]](strength)
        self.strength_scheduler = strength
        self.reset(seed)

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the strength scheduler requires to be stepped
        return None if isinstance(self.strength_scheduler, NoScheduling) else self._hook

    def reset(self, seed: RngType = None) -> None:
        self.np_random = np.random.default_rng(seed)

    def can_explore(self) -> bool:
        return True

    def step(self, *_, **__) -> None:
        """Steps (i.e., decays or increases) the exploration strength according to its
        scheduler."""
        self.strength_scheduler.step()

    def perturbation(
        self, method: str, *args: Any, **kwargs: Any
    ) -> npt.NDArray[np.floating]:
        """Returns a random perturbation.

        Parameters
        ----------
        method : str
            The name of a method from the ones available to
            :class:`numpy.random.Generator`,
            e.g., ``"random"`` for :func:`numpy.random.Generator.random`, ``"normal"``
            for :func:`numpy.random.Generator.random`, etc.
        args, kwargs
            Args and kwargs with which to call such method.

        Returns
        -------
        array
            An array representing the perturbation.
        """
        return (
            getattr(self.np_random, method)(*args, **kwargs)
            * self.strength_scheduler.value
        )

    def __repr__(self) -> str:
        stn = self.strength_scheduler.value
        return f"{self.__class__.__name__}(stn={stn},hook={self.hook},mode={self.mode})"


class EpsilonGreedyExploration(GreedyExploration):
    """Epsilon-greedy strategy for perturbing the policy, which only occasionally
    perturbs randomly the MPC policy.

    Parameters
    ----------
    epsilon : scheduler or float
        The probability to explore. Should be in range ``[0, 1]``. If passed in the form
        of an :class:`mpcrl.schedulers.Scheduler`, then the probability can be scheduled
        to decay or increase every time :meth:`step` is called. Otherwise, it is kept
        constant.
    strength : scheduler or array/supports-algebraic-operations
        The strength of the exploration. If passed in the form of an
        :class:`mpcrl.schedulers.Scheduler`, then the strength can be scheduled to decay
        or increase every time :meth:`step` is called. Otherwise, it is kept constant.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook onto, i.e., when to step the exploration's
        schedulers (if any) to, e.g., decay the chances of exploring or the perturbation
        strength (see :meth:`step` also). The options are

        - ``"on_update"``, which steps the exploration after each agent's update
        - ``"on_episode_end"``, which steps the exploration after each episode ends
        - ``"on_timestep_end"``, which steps the exploration after each env's timestep.

        By default, ``"on_update"`` is selected.
    mode : {"gradient-based", "additive"} optional
        Mode of application of explorative perturbations to the MPC. If ``"additive"``,
        then the drawn pertubation is added to the optimal action computed by the
        MPC solver. By default, ``"gradient-based"`` is selected, and in this mode the
        pertubations enter  directly in the MPC objective and is multiplied by the first
        action, thus affecting its gradient.
    seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
        Number to seed the :class:`numpy.random.Generator` used for randomizing the
        exploration. By default, ``None``.
    """

    def __init__(
        self,
        epsilon: Union[Scheduler[float], float],
        strength: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        super().__init__(strength, hook, mode, seed)
        if not isinstance(epsilon, Scheduler):
            epsilon = NoScheduling[float](epsilon)
        self.epsilon_scheduler = epsilon

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the strength or epsilon scheduler requires to be stepped
        return (
            None
            if isinstance(self.strength_scheduler, NoScheduling)
            and isinstance(self.epsilon_scheduler, NoScheduling)
            else self._hook
        )

    def can_explore(self) -> bool:
        return self.np_random.random() <= self.epsilon_scheduler.value

    def step(self, *_, **__) -> None:
        """Steps (i.e., decays or increases) the exploration strength and probability
        according to their schedulers."""
        self.strength_scheduler.step()
        self.epsilon_scheduler.step()

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        eps = self.epsilon_scheduler.value
        stn = self.strength_scheduler.value
        return f"{clsn}(eps={eps},stn={stn},hook={self.hook},mode={self.mode})"


class OrnsteinUhlenbeckExploration(ExplorationStrategy):
    """Exploration based on the Ornstein-Uhlenbeck Brownian motion with friction.

    Inspired by :class:`stable_baselines3.common.noise.OrnsteinUhlenbeckActionNoise`.

    Parameters
    ----------
    mean : scheduler or array/supports-algebraic-operations
        Mean of the stochastic process. Should have the same shape as the action.
    sigma : scheduler or array/supports-algebraic-operations
        Standard deviation of the stochastic process. Should have the same shape as
        the action.
    theta : float, optional
        Coefficient of attraction of the process towards mean, by default ``0.15``.
    dt : float, optional
        Time step of the process, by default ``1.0``.
    initial_noise : array-like, optional
        A default initial noise. By default ``None``, in which case it is set to zero.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook onto, i.e., when to step the exploration's
        schedulers (if any) to, e.g., decay the chances of exploring or the perturbation
        strength (see :meth:`step` also). The options are

        - ``"on_update"``, which steps the exploration after each agent's update
        - ``"on_episode_end"``, which steps the exploration after each episode ends
        - ``"on_timestep_end"``, which steps the exploration after each env's timestep.

        By default, ``"on_update"`` is selected.
    mode : {"gradient-based", "additive"} optional
        Mode of application of explorative perturbations to the MPC. If ``"additive"``,
        then the drawn pertubation is added to the optimal action computed by the
        MPC solver. By default, ``"gradient-based"`` is selected, and in this mode the
        pertubations enter  directly in the MPC objective and is multiplied by the first
        action, thus affecting its gradient.
    seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
        Number to seed the :class:`numpy.random.Generator` used for randomizing the
        exploration. By default, ``None``.
    """

    def __init__(
        self,
        mean: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        sigma: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        theta: float = 0.15,
        dt: float = 1.0,
        initial_noise: Optional[npt.ArrayLike] = None,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        super().__init__(hook, mode)
        if not isinstance(mean, Scheduler):
            mean = NoScheduling[npt.NDArray[np.floating]](mean)
        self.mean_scheduler = mean
        if not isinstance(sigma, Scheduler):
            sigma = NoScheduling[npt.NDArray[np.floating]](sigma)
        self.sigma_scheduler = sigma
        self.theta = theta
        self.dt = dt
        self.initial_noise = initial_noise
        self.reset(seed)

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the mean or sigma scheduler requires to be stepped
        return (
            None
            if isinstance(self.mean_scheduler, NoScheduling)
            and isinstance(self.sigma_scheduler, NoScheduling)
            else self._hook
        )

    def reset(self, seed: RngType = None) -> None:
        self.np_random = np.random.default_rng(seed)
        self._prev_noise = (
            np.zeros_like(self.mean_scheduler.value)
            if self.initial_noise is None
            else np.asarray(self.initial_noise)
        )

    def can_explore(self) -> bool:
        return True

    def step(self, *_, **__) -> None:
        """Updates (i.e., decays or increases) the mean and standard deviation of the
        perturbation according to their schedulers."""
        self.mean_scheduler.step()
        self.sigma_scheduler.step()

    def perturbation(self, *_: Any, **__: Any) -> npt.NDArray[np.floating]:
        sigma = self.sigma_scheduler.value
        noise = (
            self._prev_noise
            + (self.theta * self.dt) * (self.mean_scheduler.value - self._prev_noise)
            + np.sqrt(self.dt) * (sigma * self.np_random.normal(size=np.shape(sigma)))
        )
        self._prev_noise = noise
        return noise

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        mean = self.mean_scheduler.value
        sigma = self.sigma_scheduler.value
        return (
            f"{clsn}(mean={mean},sigma={sigma},theta={self.theta},dt={self.dt},"
            f"hook={self.hook},mode={self.mode})"
        )


class StepWiseExploration(ExplorationStrategy):
    """Wrapper-like exploration that keeps the wrapped base exploration strategy
    constants for a number of steps, thus creating a piecewise exploration.

    This class takes in another exploration instance, and allows it to change only every
    ``N`` steps, thus yielding a step-wise strategy with steps of the given length. This
    is useful when, e.g., the exploration strategy must be kept constant across time for
    a number of steps.

    Parameters
    ----------
    base_exploration : ExplorationStrategy
        The base exploration strategy to be made step-wise.
    step_size : int
        Size of each step.
    stepwise_decay : bool, optional
        Enables the decay :meth:`step` to also be step-wise, i.e., applied only every
        ``N`` steps.

    Notes
    -----
    Be carefull that this exploration wrapper ends up modifying the exploration chance
    and magnitude (if any) of the wrapped base strategy as well as the step behaviour,
    i.e., the frequency of the decay/increment of the base exploration's schedulers
    (again, if any) is enlarged by the step size factor. This is because the number of
    calls to the base exploration's :meth:`step` method is reduced by a factor of the
    step size.
    """

    def __init__(
        self,
        base_exploration: ExplorationStrategy,
        step_size: int,
        stepwise_decay: bool = True,
    ) -> None:
        super().__init__()
        del self._hook, self._mode
        self.base_exploration = base_exploration
        self.step_size = step_size
        self._explore_counter = 0
        self._step_counter = 0
        self._stepwise_decay = stepwise_decay

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        """Returns the hook of the base exploration strategy, if any."""
        return self.base_exploration.hook

    @property
    def mode(self) -> Literal["gradient-based", "additive"]:
        """Returns the mode of the base exploration strategy."""
        return self.base_exploration.mode

    def can_explore(self) -> bool:
        # since this method is called at every timestep (when deterministic=False), we
        # decide here if the base exploration is frozen or not, i.e., if we are at the
        # new step or not
        self._explore_counter %= self.step_size
        if self._explore_counter == 0:
            self._cached_can_explore = self._cached_perturbation = None
        self._explore_counter += 1

        if self._cached_can_explore is not None:
            return self._cached_can_explore
        self._cached_can_explore = self.base_exploration.can_explore()
        return self._cached_can_explore

    def step(self, *_, **__) -> None:
        if not self._stepwise_decay:
            return self.base_exploration.step()
        self._step_counter %= self.step_size
        if self._step_counter == 0:
            self.base_exploration.step()
        self._step_counter += 1

    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        if self._cached_perturbation is not None:
            return self._cached_perturbation
        self._cached_perturbation = self.base_exploration.perturbation(*args, **kwargs)
        return self._cached_perturbation

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        bclsn = self.base_exploration.__class__.__name__
        h = self.hook
        m = self.mode
        return f"{clsn}(base={bclsn},step_size={self.step_size},hook={h},mode={m})"
